#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Folding BatchNorm Graph Rewriter."""

import math

import numpy as np
from tensorflow.core.framework import attr_value_pb2, node_def_pb2
from tensorflow.python.framework import tensor_util

from neural_compressor.tensorflow.quantization.utils.graph_util import GraphAnalyzer
from neural_compressor.tensorflow.quantization.utils.graph_util import GraphRewriterHelper as Helper
from neural_compressor.tensorflow.utils import dump_elapsed_time

from ..graph_base import GraphRewriterBase


class FoldBatchNormNodesOptimizer(GraphRewriterBase):
    """Folding BatchNorm nodes into Conv."""

    INPUT_ORDER = {
        # Order of inputs for BatchNormWithGlobalNormalization.
        "BatchNormWithGlobalNormalization": ["conv_op", "mean_op", "var_op", "beta_op", "gamma_op"],
        # Order of inputs for FusedBatchNorm.
        "FusedBatchNorm": ["conv_op", "gamma_op", "beta_op", "mean_op", "var_op"],
        "FusedBatchNormV3": ["conv_op", "gamma_op", "beta_op", "mean_op", "var_op"],
        "_FusedBatchNormEx": ["conv_op", "gamma_op", "beta_op", "mean_op", "var_op"],
    }
    # Name of the attribute epsilon value is stored in.
    EPSILON_ATTR = {
        "BatchNormWithGlobalNormalization": "variance_epsilon",
        "FusedBatchNorm": "epsilon",
        "FusedBatchNormV3": "epsilon",
        "_FusedBatchNormEx": "epsilon",
    }

    def scale_after_normalization(self, node):
        """Check the scale_after_normalization attribute if the node is BatchNormWithGlobalNormalization.

        Args:
            node (nodedef): input nodedef object

        Returns:
            bool: True if the node op is not BatchNormWithGlobalNormalization else it
                    depends on the BatchNormWithGlobalNormalization attribute value of
                    `scale_after_normalization`.
        """
        if node.op == "BatchNormWithGlobalNormalization":
            return node.attr["scale_after_normalization"].b
        return True

    @dump_elapsed_time("Pass FoldBatchNormNodesOptimizer")
    def do_transformation(self):
        """Removes batch normalization ops by folding them into convolutions.

        Batch normalization during training has multiple dynamic parameters that are
        updated, but once the graph is finalized these become constants. That means
        there's an opportunity to reduce the computations down to a scale and
        addition, rather than the more expensive multiple ops, and even bake the
        scaling into the convolution weights. This function identifies the typical
        pattern of batch normalization subgraphs, and performs the transformation to
        fold the computations down into a simpler form. It currently only spots batch
        normalization that's performed by the BatchNormWithGlobalNormalization, FusedBatchNorm,
        FusedBatchNormV3 and _FusedBatchNormEx ops, and will need to be extended in the future to handle the
        newer style.

        Returns:
          Modified graph with BN ops removed, and modified weights.

        Raises:
          ValueError: If the graph is badly formed with duplicate node names.
        """
        cur_graph = GraphAnalyzer()
        cur_graph.graph = self.model

        graph_info = cur_graph.parse_graph()
        target_nodes = cur_graph.query_fusion_pattern_nodes(
            [
                ["Conv2D", "DepthwiseConv2dNative"],
                ("BiasAdd", "Add", "AddV2"),
                ["BatchNormWithGlobalNormalization", "FusedBatchNorm", "FusedBatchNormV3", "_FusedBatchNormEx"],
            ]
        )
        for node_combination in target_nodes:
            matched_node = node_combination[:-1]
            has_add_op = True if len(node_combination[-1]) == 3 else False
            conv_node = graph_info[Helper.node_name_from_input(matched_node[0])].node
            weights_node_name = graph_info[Helper.node_name_from_input(matched_node[0])].node.input[1]
            weights_node = graph_info[Helper.node_name_from_input(weights_node_name)].node
            bn_node = graph_info[Helper.node_name_from_input(matched_node[-1])].node

            # oneDNN enabled _FusedBatchNormEx only supports num_side_inputs == 0
            # and Relu/Identity activations.
            if bn_node.op == "_FusedBatchNormEx":
                if bn_node.attr["num_side_inputs"].i != 0:
                    continue
                if not (
                    bn_node.attr["activation_mode"].s == b"Identity" or bn_node.attr["activation_mode"].s == b"Relu"
                ):
                    continue

            if weights_node.op != "Const":
                self.logger.warning(
                    "Didn't find expected conv Constant input to '%s', "
                    "found %s instead. Maybe freeze_graph wasn't "
                    "run first?" % (bn_node.name, weights_node_name)
                )
                continue
            weights = Helper.values_from_const(weights_node)

            if conv_node.op == "Conv2D":
                channel_count = weights.shape[3]
            elif conv_node.op == "DepthwiseConv2dNative":
                channel_count = weights.shape[2] * weights.shape[3]

            mean_node_name = Helper.node_name_from_input(bn_node.input[self.INPUT_ORDER[bn_node.op].index("mean_op")])
            mean_node = graph_info[mean_node_name].node

            if mean_node.op != "Const":
                continue

            mean_value = Helper.values_from_const(mean_node)

            if has_add_op:
                bias_node_name = graph_info[Helper.node_name_from_input(matched_node[1])].node.input[1]
                bias_node = graph_info[Helper.node_name_from_input(bias_node_name)].node
                if bias_node.op != "Const":
                    continue

                if mean_value.shape != (channel_count,):
                    continue

                mean_value = mean_value - Helper.values_from_const(bias_node)
                cur_graph.remove_node(bias_node.name)
                cur_graph.remove_node(matched_node[1])

            if mean_value.shape != (channel_count,):
                self.logger.warning(
                    "Incorrect shape for mean, found {}, expected {}, "
                    "for node {}.".format(str(mean_value.shape), str((channel_count,)), conv_node.name)
                )
                continue
            var_node_name = Helper.node_name_from_input(bn_node.input[self.INPUT_ORDER[bn_node.op].index("var_op")])
            var_node = graph_info[var_node_name].node
            if var_node.op != "Const":
                continue
            var_value = Helper.values_from_const(var_node)

            if var_value.shape != (channel_count,):
                continue

            beta_node_name = Helper.node_name_from_input(bn_node.input[self.INPUT_ORDER[bn_node.op].index("beta_op")])
            beta_node = graph_info[beta_node_name].node
            if beta_node.op != "Const":
                continue
            beta_value = Helper.values_from_const(beta_node)

            if beta_value.shape != (channel_count,):
                continue

            gamma_node_name = Helper.node_name_from_input(bn_node.input[self.INPUT_ORDER[bn_node.op].index("gamma_op")])
            gamma_node = graph_info[gamma_node_name].node

            if gamma_node.op != "Const":
                continue
            gamma_value = Helper.values_from_const(gamma_node)

            if gamma_value.shape != (channel_count,):
                continue

            variance_epsilon_value = bn_node.attr[self.EPSILON_ATTR[bn_node.op]].f

            if self.scale_after_normalization(bn_node):
                scale_value = (1.0 / np.vectorize(math.sqrt)(var_value + variance_epsilon_value)) * gamma_value
            else:
                scale_value = 1.0 / np.vectorize(math.sqrt)(var_value + variance_epsilon_value)

            offset_value = (-mean_value * scale_value) + beta_value

            if conv_node.op == "Conv2D":
                original_shape = weights.shape
                tmp_shape = (original_shape[-1], int(weights.size / original_shape[-1]))
                tmp_order = [weights.ndim - 1] + [i for i in range(weights.ndim - 1)]
                scaled_weights = np.copy(weights).transpose(tmp_order).ravel().reshape(tmp_shape)
                reshape_scale = np.array(scale_value).reshape(len(scale_value), 1)
                scaled_weights = np.multiply(scaled_weights, reshape_scale).transpose().reshape(original_shape)
            elif conv_node.op == "DepthwiseConv2dNative":
                scaled_weights = np.copy(weights)
                it = np.nditer(scaled_weights, flags=["multi_index"], op_flags=["readwrite"])
                channel_multiplier = weights.shape[3]
                while not it.finished:
                    current_scale = scale_value[it.multi_index[2] * channel_multiplier + it.multi_index[3]]
                    it[0] *= current_scale
                    it.iternext()

            scaled_weights_node = node_def_pb2.NodeDef()
            scaled_weights_node.op = "Const"
            scaled_weights_node.name = weights_node_name + "_bn_offset"
            scaled_weights_node.attr["dtype"].CopyFrom(weights_node.attr["dtype"])
            scaled_weights_node.attr["value"].CopyFrom(
                attr_value_pb2.AttrValue(
                    tensor=tensor_util.make_tensor_proto(scaled_weights, weights.dtype.type, weights.shape)
                )
            )
            cur_graph.replace_const_node(scaled_weights_node, [conv_node.name], weights_node_name)

            offset_node = node_def_pb2.NodeDef()
            offset_node.op = "Const"
            offset_node.name = conv_node.name + "_bn_offset"
            offset_node.attr["dtype"].CopyFrom(mean_node.attr["dtype"])
            offset_node.attr["value"].CopyFrom(
                attr_value_pb2.AttrValue(
                    tensor=tensor_util.make_tensor_proto(offset_value, mean_value.dtype.type, offset_value.shape)
                )
            )
            bias_add_node = node_def_pb2.NodeDef()
            bias_add_node.op = "BiasAdd"
            bias_add_node.name = bn_node.name
            bias_add_node.attr["T"].CopyFrom(conv_node.attr["T"])
            bias_add_node.attr["data_format"].CopyFrom(conv_node.attr["data_format"])
            bias_add_node.input.extend([conv_node.name, offset_node.name])
            if bn_node.op == "_FusedBatchNormEx" and bn_node.attr["activation_mode"].s == b"Relu":
                # Create Relu op which takes Bias-Add as input.
                #  Conv2D/Depthwise-Conv2D                         Conv2D/Depthwise-Conv2D
                #      |                                                 |
                #  Bias-Add (originally _FusedBatchNormEx) <---->     Bias-Add
                #      |                                              |       \
                #   <some-node>                                  <some-node>   Relu
                relu_node = node_def_pb2.NodeDef()
                relu_node.op = "Relu"
                relu_node.name = bn_node.name + "_bn_relu"
                relu_node.attr["T"].CopyFrom(conv_node.attr["T"])
                relu_node.input.extend([bias_add_node.name])

            cur_graph.add_node(offset_node, [], [bias_add_node.name])
            cur_graph.add_node(
                bias_add_node, conv_node.name, graph_info[Helper.node_name_from_input(matched_node[-1])].outputs
            )
            if bn_node.op == "_FusedBatchNormEx" and bn_node.attr["activation_mode"].s == b"Relu":
                matchd_node_outputs = graph_info[Helper.node_name_from_input(matched_node[-1])].outputs
                cur_graph.add_node(offset_node, [], [bias_add_node.name])
                cur_graph.add_node(bias_add_node, conv_node.name, [relu_node.name])
                cur_graph.add_node(relu_node, bias_add_node.name, matchd_node_outputs)
            else:
                cur_graph.add_node(offset_node, [], [bias_add_node.name])
                cur_graph.add_node(
                    bias_add_node, conv_node.name, graph_info[Helper.node_name_from_input(matched_node[-1])].outputs
                )
            cur_graph.replace_const_node(scaled_weights_node, [conv_node.name], weights_node_name)

            cur_graph.remove_node(weights_node_name)
            cur_graph.remove_node(mean_node_name)
            cur_graph.remove_node(var_node_name)
            cur_graph.remove_node(beta_node_name)
            cur_graph.remove_node(gamma_node_name)

        return cur_graph.dump_graph()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Dequantize Cast Graph Rerewriter."""

import tensorflow as tf
from tensorflow.core.framework import attr_value_pb2
from tensorflow.python.framework import dtypes

from neural_compressor.tensorflow.quantization.utils.graph_rewriter.graph_base import GraphRewriterBase
from neural_compressor.tensorflow.quantization.utils.graph_util import GraphAnalyzer
from neural_compressor.tensorflow.utils import SPR_BASE_VERSIONS, dump_elapsed_time


class DequantizeCastOptimizer(GraphRewriterBase):
    """Remove the Cast OP and set Dequantize output to B16 if the Cast OP output is BF16."""

    @dump_elapsed_time("Pass DequantizeCastOptimizer")
    def do_transformation(self):
        """Remove the redundant Cast OP after Dequantize.

        Args:
            input_graph_def (graphdef): graphdef object

        Returns:
            [graphdef]: optimized graph
        """
        # stock TF _MklDequantize doesn't support BF16 currently.
        # TODO remove this when spr-base upstream to stock TF.
        if tf.version.VERSION not in SPR_BASE_VERSIONS:
            return self.model

        DT_BFLOAT16 = attr_value_pb2.AttrValue(type=dtypes.bfloat16.as_datatype_enum)
        cur_graph = GraphAnalyzer()
        cur_graph.graph = self.model
        graph_info = cur_graph.parse_graph()
        target_nodes = cur_graph.query_fusion_pattern_nodes([["Dequantize"], ["Cast"]])
        for i in target_nodes:
            dq_node = graph_info[i[0]].node
            dq_outputs = graph_info[i[0]].outputs

            if len(dq_outputs) > 1:
                continue
            cast_node = graph_info[i[1]].node
            cast_outputs = graph_info[i[1]].outputs
            all_cast_outputs_bf16 = True
            for cast_output in cast_outputs:
                cast_output_node = graph_info[cast_output].node
                if cast_output_node.attr["T"] != DT_BFLOAT16:  # des dtype of the cast must be bfloat16
                    all_cast_outputs_bf16 = False
            if not all_cast_outputs_bf16:
                continue
            dq_node.attr["dtype"].CopyFrom(DT_BFLOAT16)
            for cast_output in cast_outputs:
                successor_node = graph_info[cast_output].node
                replace_index = None
                for index, value in enumerate(successor_node.input):
                    if value == cast_node.name:
                        replace_index = index
                        break
                successor_node.input[replace_index] = dq_node.name

            # remove Cast node
            cur_graph.remove_node(cast_node.name)

        return cur_graph.dump_graph()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Graph Rewrite Base Class."""

import logging
from abc import abstractmethod


class GraphRewriterBase:
    """Graph Rewrite Base class.

    We abstract this base class and define the interface only.

    Args:
        object (model): the input model to be converted.
    """

    def __init__(self, model):
        """Initialization."""
        self.model = model
        self.logger = logging.getLogger("neural_compressor")

    @abstractmethod
    def do_transformation(self):
        """Base Interface that need to be implemented by each sub class."""
        raise NotImplementedError



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Strip unused nodes Graph Rewriter."""

from neural_compressor.tensorflow.utils import dump_elapsed_time

from ..graph_base import GraphRewriterBase


class StripUnusedNodesOptimizer(GraphRewriterBase):
    """Remove the unused nodes in the graph."""

    def __init__(self, model, input_node_names, output_node_names):
        """Initialization."""
        super().__init__(model)
        self.input_node_names = input_node_names
        self.output_node_names = output_node_names

    @dump_elapsed_time("Pass StripUnusedNodesOptimizer")
    def do_transformation(self):
        """Execute stripping unused nodes."""
        from neural_compressor.tensorflow.quantization.utils.utility import (
            fix_ref_type_of_graph_def,
            strip_unused_nodes,
        )

        self.model = fix_ref_type_of_graph_def(self.model)
        return strip_unused_nodes(self.model, self.input_node_names, self.output_node_names)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Fuse Transpose and Reshape Graph Rewriter."""

from tensorflow.python.framework import dtypes, tensor_util

from neural_compressor.tensorflow.quantization.utils.graph_util import GraphAnalyzer
from neural_compressor.tensorflow.quantization.utils.graph_util import GraphRewriterHelper as Helper
from neural_compressor.tensorflow.utils import dump_elapsed_time

from ..graph_base import GraphRewriterBase


class FuseTransposeReshapeOptimizer(GraphRewriterBase):
    """Fuse Transpose + Reshape + MatMul/Conv ==> MatMul/Conv."""

    @dump_elapsed_time("Pass FuseTransposeReshapeOptimizer")
    def do_transformation(self):
        """Execute Transpose + Reshape + MatMul/Conv fusion."""
        g = GraphAnalyzer()
        g.graph = self.model
        graph_info = g.parse_graph()

        patterns = [
            ["Transpose"],
            ["Reshape"],
            ["MatMul", "DepthwiseConv2dNative", "Conv2D", "BatchMatMul", "BatchMatMulV2"],
        ]

        matched_nodes = g.query_fusion_pattern_nodes(patterns)

        valid_match = []
        for i in matched_nodes:
            transpose_input_node_name = graph_info[i[0]].node.input[0]
            transpose_input_node = graph_info[transpose_input_node_name].node
            if transpose_input_node.op == "Const":
                valid_match.append(i)
            elif transpose_input_node.op == "Enter":
                if transpose_input_node.input:
                    enter_input_node_name = transpose_input_node.input[0]
                    enter_input_node = graph_info[enter_input_node_name].node
                    if enter_input_node.op == "Const":
                        valid_match.append(i)
            else:
                continue

        for i in valid_match:
            transpose_node = graph_info[i[0]].node
            transpose_input_node = graph_info[transpose_node.input[0]].node
            transpose_input_perm = graph_info[transpose_node.input[1]].node
            reshape_node = graph_info[i[1]].node
            reshape_shape_node = graph_info[reshape_node.input[1]].node
            if transpose_input_node.op == "Const":
                transpose_input_node_content = tensor_util.MakeNdarray(transpose_input_node.attr["value"].tensor)
            elif transpose_input_node.op == "Enter":
                enter_input_node_name = transpose_input_node.input[0]
                enter_input_node = graph_info[enter_input_node_name].node
                transpose_input_node_content = tensor_util.MakeNdarray(enter_input_node.attr["value"].tensor)
            else:
                continue

            if transpose_input_perm.op == "Const":
                transpose_perm_node_content = tensor_util.MakeNdarray(transpose_input_perm.attr["value"].tensor)
            elif transpose_input_perm.op == "Enter":
                enter_transpose_input_perm = transpose_input_perm.input[0]
                enter_transpose_input_perm_node = graph_info[enter_transpose_input_perm].node
                transpose_perm_node_content = tensor_util.MakeNdarray(
                    enter_transpose_input_perm_node.attr["value"].tensor
                )
            else:
                continue

            if reshape_shape_node.op == "Const":
                reshape_shape_node_content = tensor_util.MakeNdarray(reshape_shape_node.attr["value"].tensor)
            elif reshape_shape_node.op == "Enter":
                enter_reshape_shape = reshape_shape_node.input[0]
                enter_reshape_shape_node = graph_info[enter_reshape_shape].node
                reshape_shape_node_content = tensor_util.MakeNdarray(enter_reshape_shape_node.attr["value"].tensor)
            else:
                continue

            converted_node = transpose_input_node_content.transpose(transpose_perm_node_content).reshape(
                reshape_shape_node_content
            )
            g.remove_node(i[0])
            if transpose_input_node.op == "Const":
                g.remove_node(transpose_input_node.name)
                g.remove_node(transpose_input_perm.name)
                new_node_name = transpose_input_node.name + "_converted"
                new_node = Helper.create_constant_node(
                    new_node_name, converted_node, dtype=dtypes.float32, shape=converted_node.shape
                )
                g.replace_const_node(new_node, [i[2]], i[1])
                g.remove_node(i[1])
                g.remove_node(reshape_shape_node.name)
            else:
                g.remove_node(enter_input_node.name)
                g.remove_node(transpose_input_perm.name)
                new_node_name = enter_input_node.name + "_converted"

                new_node = Helper.create_constant_node(
                    new_node_name, converted_node, dtype=dtypes.float32, shape=converted_node.shape
                )
                g.add_node(new_node, [], [transpose_input_node.name])
                transpose_input_node.input[0] = new_node.name
                for index, node_name in enumerate(graph_info[i[2]].node.input):
                    if node_name == i[1]:
                        graph_info[i[2]].node.input[index] = transpose_input_node.name
                g.remove_node(i[1])
                g.remove_node(reshape_shape_node.name)

        return g.dump_graph()



#
#  -*- coding: utf-8 -*-
#
#  Copyright (c) 2021 Intel Corporation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
"""Graph rewriter BF16 Converter Class."""

from __future__ import absolute_import, division, print_function

import copy
import logging

import tensorflow as tf
from tensorflow.core.framework import attr_value_pb2
from tensorflow.python.framework import dtypes, op_def_registry, tensor_util
from tensorflow.python.framework.kernels import get_registered_kernels_for_op

from neural_compressor.tensorflow.quantization.utils.graph_rewriter.bf16.dequantize_cast_optimizer import (
    DequantizeCastOptimizer,
)
from neural_compressor.tensorflow.quantization.utils.graph_rewriter.generic.graph_cse_optimizer import GraphCseOptimizer
from neural_compressor.tensorflow.quantization.utils.graph_rewriter.graph_base import GraphRewriterBase
from neural_compressor.tensorflow.quantization.utils.graph_util import GraphAnalyzer
from neural_compressor.tensorflow.quantization.utils.graph_util import GraphRewriterHelper as Helper
from neural_compressor.tensorflow.utils import SPR_BASE_VERSIONS

DT_FLOAT32 = attr_value_pb2.AttrValue(type=dtypes.float32.as_datatype_enum)
DT_BFLOAT16 = attr_value_pb2.AttrValue(type=dtypes.bfloat16.as_datatype_enum)


class BF16Convert(GraphRewriterBase):
    """BF16 node convert transformation."""

    def __init__(self, model, fp32_ops=[], bf16_ops=[]):
        """Initialization.

        Args: model: the model to be converted to BF16.
              fp32_ops: keep with fp32 op list
              bf16_ops: convert to bf16 op list
        """
        super().__init__(model)

        self.cur_graph = GraphAnalyzer()
        self.cur_graph.graph = self.model
        self.fp32_ops = fp32_ops
        self.bf16_ops = bf16_ops
        self.converted_ops = []
        self.device = ["CPU", "DEFAULT"]  # TODO support different device types, such as GPU

    def _dtype(self, node):
        """Get the dtype of the node."""
        op_def = op_def_registry.get(node.op)
        inputs_dt = []
        outputs_dt = []
        for i in op_def.input_arg:
            inputs_num = node.attr[i.number_attr].i if i.number_attr else 1
            for j in range(inputs_num):
                if i.type:
                    inputs_dt.append("")
                else:
                    inputs_dt.append(i.type_attr)
        for i in op_def.output_arg:
            outputs_num = node.attr[i.number_attr].i if i.number_attr else 1
            for j in range(outputs_num):
                if i.type:
                    outputs_dt.append("")
                else:
                    outputs_dt.append(i.type_attr)
        return inputs_dt, outputs_dt

    def _dtype_val(self, node):
        """Get the dtype value of the node."""
        op_def = op_def_registry.get(node.op)
        inputs_dt_val = []
        outputs_dt_val = []
        for i in op_def.input_arg:
            inputs_num = node.attr[i.number_attr].i if i.number_attr else 1
            for j in range(inputs_num):
                if i.type:
                    inputs_dt_val.append(copy.deepcopy(attr_value_pb2.AttrValue(type=i.type)))
                else:
                    inputs_dt_val.append(copy.deepcopy(node.attr[i.type_attr]))
        for i in op_def.output_arg:
            outputs_num = node.attr[i.number_attr].i if i.number_attr else 1
            for j in range(outputs_num):
                if i.type:
                    outputs_dt_val.append(copy.deepcopy(attr_value_pb2.AttrValue(type=i.type)))
                else:
                    outputs_dt_val.append(copy.deepcopy(node.attr[i.type_attr]))
        return inputs_dt_val, outputs_dt_val

    def _allowed_dtype_val(self, node):
        """Get the allowed dtype value of the node."""
        op_def = op_def_registry.get(node.op)
        allowed_dt_val = {}
        for attr_def in op_def.attr:
            if attr_def.type != "type":
                continue
            if attr_def.HasField("allowed_values"):
                allowed_dt_val[attr_def.name] = attr_def.allowed_values.list.type
        # The supported data type in op_def may be different with registered kernels.
        # Use the registered one if exists.
        registered_dt_val = {}
        registered_kernels = get_registered_kernels_for_op(node.op)
        for kernel in registered_kernels.kernel:
            if kernel.device_type in self.device:
                for constraint in kernel.constraint:
                    if constraint.HasField("allowed_values"):
                        if constraint.name not in registered_dt_val:
                            registered_dt_val[constraint.name] = constraint.allowed_values.list.type
                        else:
                            registered_dt_val[constraint.name].extend(constraint.allowed_values.list.type)
        for dt_val in registered_dt_val:
            if registered_dt_val[dt_val] != []:
                allowed_dt_val[dt_val] = registered_dt_val[dt_val]
        return allowed_dt_val

    def _bf16_convert(self, bf16_node_name):
        """BF16 conversion for the model.

        Args: bf16_node_name: nodes converted to BF16 op list
        """
        bf16_node_detail = self.cur_graph.node_name_details[bf16_node_name]
        bf16_node = bf16_node_detail.node
        bf16_node_outputs = copy.deepcopy(bf16_node_detail.outputs)

        if bf16_node.name in self.converted_ops:
            return
        elif "Dequantize" in bf16_node.op:
            return
        else:
            self.converted_ops.append(bf16_node.name)

        inputs_dt, outputs_dt = self._dtype(bf16_node)
        inputs_dt_val, outputs_dt_val = self._dtype_val(bf16_node)
        allowed_dt_val = self._allowed_dtype_val(bf16_node)
        for index, input_name in enumerate(bf16_node.input):
            if input_name.startswith("^"):
                continue

            input_detail = self.cur_graph.node_name_details[Helper.node_name_from_input(input_name)]
            input_node = input_detail.node
            input_node_outputs = input_detail.outputs
            if (
                inputs_dt[index] in allowed_dt_val
                and dtypes.bfloat16.as_datatype_enum not in allowed_dt_val[inputs_dt[index]]
            ):
                continue

            if inputs_dt_val[index] != DT_FLOAT32:
                continue
            if (
                input_node.op == "Cast"
                and input_node.attr["SrcT"] == DT_BFLOAT16
                and input_node.attr["DstT"] == DT_FLOAT32
                and len(input_node_outputs) == 1
            ):
                parent_input_name = Helper.node_name_from_input(input_node.input[0])
                bf16_node.input[index] = input_node.input[0]
                outputs = self.cur_graph.node_name_details[parent_input_name].outputs
                outputs = list(map(lambda x: x.replace(input_name, bf16_node.name), outputs))
                self.cur_graph.remove_node(input_name)
            elif input_node.op == "Cast" and input_node.attr["DstT"] == DT_FLOAT32 and len(input_node_outputs) == 1:
                input_node.attr["DstT"].CopyFrom(DT_BFLOAT16)
            elif input_node.op == "Const" and len(input_node_outputs) == 1:
                fp32_value = tensor_util.MakeNdarray(input_node.attr.get("value").tensor)
                Helper.set_attr_dtype(input_node, "dtype", dtypes.bfloat16)
                input_node.attr["value"].CopyFrom(
                    attr_value_pb2.AttrValue(
                        tensor=tensor_util.make_tensor_proto(fp32_value, dtypes.bfloat16, fp32_value.shape)
                    )
                )
            elif (
                "Dequantize" == input_node.op
                and len(input_node_outputs) == 1
                and input_node.attr["mode"].s != b"MIN_FIRST"
                and tf.version.VERSION in SPR_BASE_VERSIONS
            ):
                # Dequantize with mode MIN_FIRST does not support bf16 in both eigen and mkl
                _, outputs_dt_input_node = self._dtype(input_node)
                allowed_input_node_dt_val = self._allowed_dtype_val(input_node)
                if (
                    outputs_dt_input_node[0] in allowed_input_node_dt_val
                    and dtypes.bfloat16.as_datatype_enum in allowed_input_node_dt_val[outputs_dt_input_node[0]]
                ):
                    input_node.attr[outputs_dt_input_node[0]].CopyFrom(DT_BFLOAT16)
            # ResizeBilinear input can be of different types but output is always float
            elif (
                input_node.name in self.bf16_ops
                and "Dequantize" not in input_node.op
                and input_node.op != "ResizeBilinear"
            ):
                self._bf16_convert(input_node.name)
            else:
                cast_node_name = input_name.replace(":", "_") + "/" + bf16_node_name + "_FP32toBF16"
                if cast_node_name not in list(self.cur_graph.node_name_details.keys()):
                    input_cast_node = Helper.create_node("Cast", cast_node_name, [input_name])
                    Helper.set_attr_dtype(input_cast_node, "DstT", dtypes.bfloat16)
                    Helper.set_attr_dtype(input_cast_node, "SrcT", dtypes.float32)
                    Helper.set_attr_bool(input_cast_node, "Truncate", False)
                bf16_node.input[index] = cast_node_name
                outputs = self.cur_graph.node_name_details[Helper.node_name_from_input(input_name)].outputs
                outputs = list(map(lambda x: x.replace(bf16_node.name, cast_node_name), outputs))
                self.cur_graph.add_node(input_cast_node, input_name, [bf16_node_name])

            bf16_node.attr[inputs_dt[index]].CopyFrom(attr_value_pb2.AttrValue(type=dtypes.bfloat16.as_datatype_enum))

        for output_name in bf16_node_outputs:
            if bf16_node.op == "ResizeBilinear":
                continue
            output_detail = self.cur_graph.node_name_details[output_name]
            output_node = output_detail.node
            inputs_dt_input_node, _ = self._dtype(output_node)
            allowed_output_node_dt_val = self._allowed_dtype_val(output_node)

            for i, input_name in enumerate(output_node.input):
                if input_name.startswith("^"):
                    continue

                if bf16_node.name != input_name.split(":")[0]:
                    continue

                index = int(input_name.split(":")[-1]) if ":" in input_name else 0
                if (
                    outputs_dt[index] in allowed_dt_val
                    and dtypes.bfloat16.as_datatype_enum not in allowed_dt_val[outputs_dt[index]]
                ):
                    continue
                if outputs_dt_val[index] != DT_FLOAT32:
                    continue

                if output_node.op == "Cast":
                    output_node.attr["SrcT"].CopyFrom(DT_BFLOAT16)
                elif output_node.op == "QuantizeV2" and "dtype" in output_node.attr:
                    if (
                        "dtype" in allowed_output_node_dt_val
                        and dtypes.bfloat16.as_datatype_enum in allowed_output_node_dt_val["dtype"]
                    ):
                        output_node.attr["dtype"].CopyFrom(DT_BFLOAT16)
                elif (
                    output_node.name not in self.bf16_ops
                    or inputs_dt_input_node[i] in allowed_output_node_dt_val
                    and dtypes.bfloat16.as_datatype_enum not in allowed_output_node_dt_val[inputs_dt_input_node[i]]
                ):
                    cast_node_name = bf16_node_name + "/" + output_node.name + "_BF16toFP32"
                    if cast_node_name in self.cur_graph.node_name_details.keys():
                        continue
                    output_cast_node = Helper.create_node("Cast", cast_node_name, [input_name])
                    Helper.set_attr_dtype(output_cast_node, "DstT", dtypes.float32)
                    Helper.set_attr_dtype(output_cast_node, "SrcT", dtypes.bfloat16)
                    Helper.set_attr_bool(output_cast_node, "Truncate", False)
                    index = [i for i in output_node.input].index(input_name)
                    output_node.input[index] = output_cast_node.name
                    self.cur_graph.add_node(output_cast_node, bf16_node_name, [output_name])

    def _model_bf16_convert(self):
        """Convert model to BF16."""
        logging.debug("start convert bf16 graph")
        self.cur_graph.parse_graph()
        for bf16_node_name in set(self.bf16_ops):
            if bf16_node_name not in self.cur_graph.node_name_details:
                self.bf16_ops.remove(bf16_node_name)
        for bf16_node_name in sorted(list(set(self.bf16_ops))):
            self._bf16_convert(bf16_node_name)
        return self.cur_graph.dump_graph()

    def do_transformation(self):
        """Execute BF16 convert.

        Returns: Transformed graph
        """
        converted_graph_def = self._model_bf16_convert()
        # remove those ops which could be shared by Graph Cse optimizer
        converted_graph_def = GraphCseOptimizer(converted_graph_def).do_transformation()
        # remove cast and set dequantize dtype bf16 when all outputs of dequantize are bf16
        converted_graph_def = DequantizeCastOptimizer(converted_graph_def).do_transformation()
        converted_graph_def.library.CopyFrom(self.model.library)
        return converted_graph_def



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Folding Const Graph Rewriter."""


import numpy as np
import tensorflow as tf
from tensorflow.python.platform import tf_logging

from neural_compressor.tensorflow.quantization.utils.graph_util import GraphAnalyzer, GraphRewriterHelper
from neural_compressor.tensorflow.utils import dump_elapsed_time

from ..graph_base import GraphRewriterBase


class GraphFoldConstantOptimizer(GraphRewriterBase):
    """Folding all the sequences only consist of const and self.supported_op_type."""

    supported_op_type = ["Add", "AddV2", "Const", "Mul", "Rsqrt", "Sub"]

    def __init__(self, model=None):
        """Initialization."""
        super().__init__(model)
        self.graph_analyzer = GraphAnalyzer()
        self.graph_analyzer.graph = self.model

        self.graph_info = self.graph_analyzer.parse_graph()

    def _fold_value(self, end_node_name):
        """Calculate values of end node of constant node sequence.

        there may be layers whose inputs are all constant in the graph, like:
          const
                > add
          const
        the value of add can be calculated in advance.

        Args:
          end_node_name: name of the end node of the sequence. e.g. add in the above examples.

        Returns:
          values of end node.

        Raises:
          ValueError: If the graph contains tensors which can't be broadcast.
        """
        end_node = self.graph_info[end_node_name].node

        def can_broadcast(s1, s2):
            if s1.shape and s2.shape:
                s1a = np.asarray(s1.shape)
                s2a = np.asarray(s2.shape)
                return ((s1a == 1) | (s2a == 1) | (s2a == s1a)).all()

            return True

        if self.graph_info[end_node_name].node.input:
            if end_node.op == "Mul":
                first_value = self._fold_value(list(end_node.input)[0])
                first_type = first_value.dtype
                fold_value = np.array(1.0).astype(first_type)
                for index, input in enumerate(end_node.input):
                    # broadcast if needed
                    input_value = self._fold_value(input)
                    input_type = input_value.dtype
                    if can_broadcast(fold_value, input_value):
                        fold_value = fold_value * input_value
                    else:
                        raise ValueError("input {} of node {} can't be broadcast".format(input.name, end_node.name))
                return fold_value.astype(first_type)
            elif end_node.op == "Add" or end_node.op == "AddV2":
                first_value = self._fold_value(list(end_node.input)[0])
                first_type = first_value.dtype
                fold_value = np.array(0.0).astype(first_type).reshape(())
                for index, input in enumerate(end_node.input):
                    # broadcast if needed
                    input_value = self._fold_value(input)
                    if can_broadcast(fold_value, input_value):
                        fold_value = fold_value + input_value
                    else:
                        raise ValueError("input {} of node {} can't be broadcast".format(input.name, end_node.name))
                return fold_value.astype(first_type)
            elif end_node.op == "Rsqrt":
                return 1 / np.sqrt(self._fold_value(end_node.input[0]))
            elif end_node.op == "Sub":
                first_value = self._fold_value(list(end_node.input)[0])
                first_type = first_value.dtype
                fold_value = np.array(0.0, dtype=first_type)
                for index, input in enumerate(end_node.input):
                    # broadcast if needed
                    input_value = self._fold_value(input)
                    if first_type != input_value.dtype:
                        raise ValueError(
                            "input of node {} must be in same dtype but get {}and {}".format(
                                input.name, first_type, input_value.dtype
                            )
                        )
                    if can_broadcast(fold_value, input_value):
                        fold_value = fold_value + (-1) ** index * input_value
                    else:
                        raise ValueError("input {} of node {} can't be broadcast".format(input.name, end_node.name))
                return fold_value.astype(first_type)
            else:
                tf_logging.info(
                    "Currently fold-constant only support limited ops {} but face {}".format(
                        self.supported_op_type, end_node.op
                    )
                )
        else:
            return GraphRewriterHelper.values_from_const(end_node)

    def check_all_folded(self):
        """Check the node has been folded completely.

        Returns:
            bool: True if the node has been folded else False.
        """
        for node_name, _ in self.graph_info.items():
            if self.check_const_inputs(node_name):
                return False
        return True

    def check_const_inputs(self, node_name):
        """Check the node has the const input.

        Args:
            node_name (string): node name

        Returns:
            bool: True if the node has the const input else False
        """
        if node_name not in self.graph_info:
            return False
        node_op = self.graph_info[node_name].node.op
        if node_op == "Placeholder" or node_op == "Const":
            return False
        if node_op not in self.supported_op_type:
            return False
        constant_flag = True
        for input_name in self.graph_info[node_name].node.input:
            input_name = GraphRewriterHelper.node_name_from_input(input_name)
            input_node = self.graph_info[input_name].node
            constant_flag &= input_node.op == "Const" and not input_node.input
        return constant_flag

    @dump_elapsed_time("Pass GraphFoldConstantOptimizer")
    def do_transformation(self):
        """Fold all the sequences only consist of const and self.supported_op_type.

        Args:
          input_graph_def (graphdef): graphdef object

        Returns:
           [graphdef]: optimized graph
        """
        while not self.check_all_folded():
            for node_name, _ in self.graph_info.copy().items():
                if self.check_const_inputs(node_name):
                    fold_value = self._fold_value(node_name)
                    fold_type = tf.as_dtype(fold_value.dtype)
                    new_constant_node = GraphRewriterHelper.create_constant_node(
                        node_name + "_const", fold_value, fold_type
                    )
                    self.graph_analyzer.replace_constant_graph_with_constant_node(new_constant_node, node_name)

        output_graph_def = self.graph_analyzer.dump_graph()

        return output_graph_def



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Fuse QuantizedMatMul with redundant Dequantize Graph Rewriter."""

from tensorflow.core.framework import attr_value_pb2, node_def_pb2
from tensorflow.python.framework import dtypes

from neural_compressor.tensorflow.quantization.utils.graph_util import GraphAnalyzer
from neural_compressor.tensorflow.quantization.utils.graph_util import GraphRewriterHelper as Helper

from ..graph_base import GraphRewriterBase


class FuseMatMulRedundantDequantizeTransformer(GraphRewriterBase):
    """Fuse _QuantizedMatMul with the successor Dequantize Op."""

    fuse_patterns = [["_QuantizedMatMul", "_QuantizedBatchMatMul"], ["Dequantize", "Cast"]]

    def __init__(self, model, device="cpu"):
        """Initialization."""
        super().__init__(model)
        self.device = device
        self.graph_analyzer = GraphAnalyzer()
        self.graph_analyzer.graph = self.model
        self.graph_info = self.graph_analyzer.parse_graph()

    def do_transformation(self):
        """Fuse the _QuantizedMatMul with the following Dequantize op.

        The output of _QuantizedMatMul or is fp32 or bf16.

        Returns:
            [graphdef]: the optimized graphdef object
        """
        target_nodes = self.graph_analyzer.query_fusion_pattern_nodes(self.fuse_patterns)

        for i in target_nodes:
            quantized_node_name = i[0]
            quantized_node = self.graph_info[quantized_node_name].node
            dequantize_node_name = i[1]
            dequantize_node = self.graph_info[dequantize_node_name].node

            if len(self.graph_info[quantized_node_name].outputs) > 3:  # pragma: no cover
                need_drop = False
                for output in self.graph_info[quantized_node_name].outputs:
                    if self.graph_info[output].node.op != "Dequantize":
                        need_drop = True
                        break
                if need_drop:
                    continue

            # ignore shared output case for license-plate-recognition-barrier-0007 model
            if (
                len(self.graph_info[dequantize_node_name].outputs) == 2
                and self.graph_info[self.graph_info[dequantize_node_name].outputs[0]].node.op == "Reshape"
                and self.graph_info[self.graph_info[dequantize_node_name].outputs[1]].node.op == "Shape"
            ):
                continue

            new_node = node_def_pb2.NodeDef()
            new_node.op = quantized_node.op

            if dequantize_node.op == "Dequantize":
                fused_ops = str(quantized_node.attr["fused_ops"].list.s).replace("Requantize", "Dequantize")
            new_node.name = quantized_node.name + "_dequantize"

            for _, value in enumerate(quantized_node.input):
                new_node.input.append(value)

            if "input_quant_mode" in quantized_node.attr:
                new_node.attr["input_quant_mode"].CopyFrom(quantized_node.attr["input_quant_mode"])
            if "output_quant_mode" in quantized_node.attr:
                new_node.attr["output_quant_mode"].CopyFrom(quantized_node.attr["output_quant_mode"])
            if "leakyrelu_alpha" in quantized_node.attr:
                new_node.attr["leakyrelu_alpha"].CopyFrom(quantized_node.attr["leakyrelu_alpha"])
            if "T1" in quantized_node.attr:
                new_node.attr["T1"].CopyFrom(quantized_node.attr["T1"])
            if "T2" in quantized_node.attr:
                new_node.attr["T2"].CopyFrom(quantized_node.attr["T2"])
            if "U" in quantized_node.attr:
                new_node.attr["U"].CopyFrom(quantized_node.attr["U"])
            if "is_weight_const" in quantized_node.attr:
                new_node.attr["is_weight_const"].CopyFrom(quantized_node.attr["is_weight_const"])
            if "is_bias_const" in quantized_node.attr:
                new_node.attr["is_bias_const"].CopyFrom(quantized_node.attr["is_bias_const"])
            if "transpose_a" in quantized_node.attr:
                new_node.attr["transpose_a"].CopyFrom(quantized_node.attr["transpose_a"])
            if "transpose_b" in quantized_node.attr:
                new_node.attr["transpose_b"].CopyFrom(quantized_node.attr["transpose_b"])
            if "Tdevice_inputs" in quantized_node.attr:
                new_node.attr["Tdevice_inputs"].CopyFrom(quantized_node.attr["Tdevice_inputs"])
            if "Tdevice_outputs" in quantized_node.attr:
                new_node.attr["Tdevice_outputs"].CopyFrom(quantized_node.attr["Tdevice_outputs"])
            if "Thost_inputs" in quantized_node.attr:
                new_node.attr["Thost_inputs"].CopyFrom(quantized_node.attr["Thost_inputs"])
            if "Tbias" in quantized_node.attr:
                new_node.attr["Tbias"].CopyFrom(quantized_node.attr["Tbias"])
            if "adj_x" in quantized_node.attr:
                new_node.attr["adj_x"].CopyFrom(quantized_node.attr["adj_x"])
            if "adj_y" in quantized_node.attr:
                new_node.attr["adj_y"].CopyFrom(quantized_node.attr["adj_y"])
            if "input_quant_mode" in quantized_node.attr:
                new_node.attr["input_quant_mode"].CopyFrom(quantized_node.attr["input_quant_mode"])
            if "output_quant_mode" in quantized_node.attr:
                new_node.attr["output_quant_mode"].CopyFrom(quantized_node.attr["output_quant_mode"])
            if "fused_ops" in quantized_node.attr:
                new_node.attr["fused_ops"].CopyFrom(quantized_node.attr["fused_ops"])

            # update Tbias for single MatMul without bias case, same as Tout.
            if dequantize_node.op == "Dequantize":
                Helper.set_attr_type_list(new_node, "Thost_outputs", [dequantize_node.attr["dtype"].type])
                new_node.attr["Tout"].CopyFrom(attr_value_pb2.AttrValue(type=dequantize_node.attr["dtype"].type))
                if new_node.op == "_QuantizedBatchMatMul":
                    new_node.attr["U"].CopyFrom(attr_value_pb2.AttrValue(type=dequantize_node.attr["DstT"].type))
                if str(quantized_node.attr["fused_ops"].list.s) == str([b"Requantize"]):
                    new_node.attr["Tbias"].CopyFrom(attr_value_pb2.AttrValue(type=dequantize_node.attr["dtype"].type))
                Helper.set_attr_string_list(new_node, "fused_ops", eval(fused_ops))
            else:
                Helper.set_attr_type_list(new_node, "Thost_outputs", [dequantize_node.attr["DstT"].type])
                new_node.attr["Tout"].CopyFrom(attr_value_pb2.AttrValue(type=dequantize_node.attr["DstT"].type))
                if new_node.op == "_QuantizedBatchMatMul":
                    new_node.attr["U"].CopyFrom(attr_value_pb2.AttrValue(type=dequantize_node.attr["DstT"].type))

            top_node_name = Helper.node_name_from_input(quantized_node.input[0])
            if self.graph_info[dequantize_node_name].outputs:
                self.graph_analyzer.replace_single_node(
                    new_node,
                    [top_node_name],
                    quantized_node_name,
                    self.graph_info[dequantize_node_name].outputs,
                    dequantize_node_name,
                )
                self.graph_analyzer.remove_node(dequantize_node_name)
            else:
                self.graph_analyzer.remove_node(dequantize_node_name)

                new_node.name = dequantize_node_name
                self.graph_analyzer.replace_single_node(
                    new_node, [top_node_name], quantized_node_name, [], dequantize_node_name
                )

            self.graph_analyzer.remove_node(quantized_node_name)

        return self.graph_analyzer.dump_graph()



#
#  -*- coding: utf-8 -*-
#
#  Copyright (c) 2021 Intel Corporation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
"""ConcatV2 rerange transform."""

from __future__ import absolute_import, division, print_function

import numpy as np
from tensorflow.core.framework import attr_value_pb2
from tensorflow.python.framework import dtypes, tensor_util

from neural_compressor.tensorflow.quantization.utils.graph_util import GraphRewriterHelper as Helper

from .graph_transform_base import GraphTransformBase


class RerangeQuantizedConcat(GraphTransformBase):
    """This class implements the rerange_quantize concat graph transform."""

    fused_requantized_bias_op = (
        "QuantizedConv2DWithBiasAndRequantize",
        "QuantizedConv2DWithBiasAndReluAndRequantize",
        "QuantizedConv2DWithBiasSumAndReluAndRequantize",
        "QuantizedConv2DWithBiasSignedSumAndReluAndRequantize",
    )
    fuse_requantized_bias_op_new_api = (
        [b"BiasAdd", b"Requantize"],
        [b"BiasAdd", b"Relu", b"Requantize"],
        [b"BiasAdd", b"LeakyRelu", b"Requantize"],
        [b"BiasAdd", b"Sum", b"Relu", b"Requantize"],
        [b"BiasAdd", b"Sum", b"Requantize"],
        [b"BiasAdd", b"Elu", b"Requantize"],
    )
    fuse_requantized_relu_op_new_api = (
        [b"BiasAdd", b"Relu", b"Requantize"],
        [b"BiasAdd", b"LeakyRelu", b"Requantize"],
        # [b'BiasAdd', b'Sum',  b'Relu',  b'Requantize']
        [b"BiasAdd", b"Elu", b"Requantize"],
    )
    offset_map = {
        "QuantizedConv2DAndRequantize": 6,
        "QuantizedConv2DAndReluAndRequantize": 6,
        "QuantizedConv2DWithBiasAndRequantize": 7,
        "QuantizedConv2DWithBiasAndReluAndRequantize": 7,
        "QuantizedConv2DWithBiasSumAndReluAndRequantize": 7,
        "QuantizedConv2DWithBiasSignedSumAndReluAndRequantize": 7,
    }
    offset_map_new_api = {
        str([b"Requantize"]): 6,
        str([b"Relu", b"Requantize"]): 6,
        str([b"LeakyRelu", b"Requantize"]): 6,
        str([b"BiasAdd", b"Requantize"]): 7,
        str([b"BiasAdd", b"Relu", b"Requantize"]): 7,
        str([b"BiasAdd", b"LeakyRelu", b"Requantize"]): 7,
        str([b"BiasAdd", b"Elu", b"Requantize"]): 7,
        str([b"BiasAdd", b"Sum", b"Relu", b"Requantize"]): 10,
        str([b"BiasAdd", b"Sum", b"Requantize"]): 10,
    }

    def __init__(self, input_pb, device, performance_only=False):
        """Initialization."""
        super().__init__(input_pb)

        self.parse_input_pb()
        self.concat_node_input_mapping = {}
        self.rerange_concat_node = []
        self.device = device
        self.performance_only = performance_only

    def _analyze_concat_node_recursively(self, quantized_conv_nodes, input_node):
        """Analyze all the ConcatV2 nodes recursively."""
        op_type = input_node.op
        if op_type == "QuantizedConcatV2":
            can_rerange = True
            concat_input_num = input_node.attr["N"].i
            for index in range(concat_input_num):
                concat_input_node = self.node_mapping[self.get_node_name_from_input(input_node.input[index])]
                concat_input_node_op_type = concat_input_node.op
                if concat_input_node_op_type in self.offset_map:
                    quantized_conv_nodes.append(concat_input_node)
                elif (
                    concat_input_node.op == "_FusedQuantizedConv2D"
                    and "fused_ops" in concat_input_node.attr
                    and str(concat_input_node.attr["fused_ops"].list.s) in self.offset_map_new_api
                ):
                    quantized_conv_nodes.append(concat_input_node)
                elif concat_input_node_op_type in ("QuantizedMaxPool", "QuantizedAvgPool"):
                    another_concat_node = self.node_mapping[self.get_node_name_from_input(concat_input_node.input[0])]
                    if self.performance_only:
                        if (
                            another_concat_node.op == "_FusedQuantizedConv2D"
                            and "fused_ops" in another_concat_node.attr
                            and str(another_concat_node.attr["fused_ops"].list.s) in self.offset_map_new_api
                        ):
                            quantized_conv_nodes.append(another_concat_node)
                    else:
                        if not self._analyze_concat_node_recursively(quantized_conv_nodes, another_concat_node):
                            can_rerange = False
                            break
                elif concat_input_node_op_type == "QuantizedConcatV2":
                    if not self._analyze_concat_node_recursively(quantized_conv_nodes, concat_input_node):
                        can_rerange = False
                        break
                elif self.performance_only and concat_input_node_op_type == "QuantizeV2":
                    quantized_conv_nodes.append(concat_input_node)
                else:
                    can_rerange = False
                    break

            return can_rerange
        elif op_type == "QuantizedConv2DWithBiasAndReluAndRequantize" or (
            input_node.op == "_FusedQuantizedConv2D"
            and "fused_ops" in input_node.attr
            and input_node.attr["fused_ops"].list.s in self.fuse_requantized_relu_op_new_api
        ):
            can_rerange = True
            quantized_conv_nodes.append(input_node)
            return can_rerange
        else:
            return False

    def do_transformation(self):
        """Apply the rerange quantized ConcatV2 transform."""
        for _, node in enumerate(self.input_graph.node):
            if node.op != "QuantizedConcatV2":
                continue
            quantized_conv_nodes = []
            can_rerange = self._analyze_concat_node_recursively(quantized_conv_nodes, node)
            if not can_rerange:
                continue

            self.rerange_concat_node.append(node.name)
            combined_min = np.finfo(np.float64).max
            combined_max = -combined_min

            for node in quantized_conv_nodes:
                offset_value = 6
                if (
                    node.op == "_FusedQuantizedConv2D"
                    and "fused_ops" in node.attr
                    and str(node.attr["fused_ops"].list.s) in self.offset_map_new_api
                ):
                    offset_value = self.offset_map_new_api[str(node.attr["fused_ops"].list.s)]
                elif self.performance_only and node.op == "QuantizeV2":
                    offset_value = 1
                else:
                    offset_value = self.offset_map[node.op]
                min_value_node = self.node_mapping[node.input[offset_value]]
                max_value_node = self.node_mapping[node.input[offset_value + 1]]
                min_value = min_value_node.attr["value"].tensor.float_val[0]
                max_value = max_value_node.attr["value"].tensor.float_val[0]
                if min_value < combined_min:
                    combined_min = min_value
                if max_value > combined_max:
                    combined_max = max_value

            if self.performance_only:
                combined_value = max(abs(combined_min), abs(combined_max))
                combined_min = -combined_value
                combined_max = combined_value

            for node in quantized_conv_nodes:
                offset_value = 6
                if (
                    node.op == "_FusedQuantizedConv2D"
                    and "fused_ops" in node.attr
                    and str(node.attr["fused_ops"].list.s) in self.offset_map_new_api
                ):
                    offset_value = self.offset_map_new_api[str(node.attr["fused_ops"].list.s)]
                elif self.performance_only and node.op == "QuantizeV2":
                    offset_value = 1
                else:
                    offset_value = self.offset_map[node.op]

                min_value_node = self.node_mapping[node.input[offset_value]]
                max_value_node = self.node_mapping[node.input[offset_value + 1]]
                min_value_node.attr["value"].CopyFrom(
                    attr_value_pb2.AttrValue(
                        tensor=tensor_util.make_tensor_proto(float(combined_min), dtypes.float32, [])
                    )
                )

                max_value_node.attr["value"].CopyFrom(
                    attr_value_pb2.AttrValue(
                        tensor=tensor_util.make_tensor_proto(float(combined_max), dtypes.float32, [])
                    )
                )
        if self.device == "cpu":
            self._update_bias()
        return self.input_graph

    def _update_bias(self):
        """Convert the bias from float to int."""
        for node_name in self.node_mapping:
            current_node = self.node_mapping[node_name]
            current_node_op = current_node.op
            if (current_node_op in self.fused_requantized_bias_op) or (
                current_node_op == "_FusedQuantizedConv2D"
                and "fused_ops" in current_node.attr
                and current_node.attr["fused_ops"].list.s in self.fuse_requantized_bias_op_new_api
            ):
                done = False
                another_conv_node = None
                original_conv_node = current_node
                while not done:
                    current_node = self.node_mapping[self.get_node_name_from_input(current_node.input[0])]
                    if current_node.op in self.offset_map:
                        another_conv_node = current_node
                        done = True
                    elif (
                        current_node.op == "_FusedQuantizedConv2D"
                        and "fused_ops" in current_node.attr
                        and str(current_node.attr["fused_ops"].list.s) in self.offset_map_new_api
                    ):
                        another_conv_node = current_node
                        done = True
                    elif current_node.op == "QuantizedConcatV2":
                        if current_node.name not in self.rerange_concat_node:
                            done = True
                    elif current_node.op not in ("QuantizedMaxPool", "QuantizedAvgPool"):
                        done = True

                if not another_conv_node:
                    continue

                bias_node = self.node_mapping[self.get_node_name_from_input(original_conv_node.input[2])]
                bias_node_type = original_conv_node.attr["Tbias"]

                if bias_node_type.type != dtypes.float32 or bias_node_type.type == dtypes.qint32:
                    continue
                sum_off_set = 0
                if original_conv_node.op == "_FusedQuantizedConv2D":
                    if str(original_conv_node.attr["fused_ops"].list.s) == str(
                        [b"BiasAdd", b"Sum", b"Relu", b"Requantize"]
                    ) or str(original_conv_node.attr["fused_ops"].list.s) == str([b"BiasAdd", b"Sum", b"Requantize"]):
                        sum_off_set = 1
                    # else:
                    #    print(str(original_conv_node.attr['fused_ops'].list.s))
                min_filter_node = self.node_mapping[original_conv_node.input[5 + sum_off_set]]
                max_filter_node = self.node_mapping[original_conv_node.input[6 + sum_off_set]]

                channel_size = (
                    1
                    if not min_filter_node.attr["value"].tensor.tensor_shape.dim
                    else min_filter_node.attr["value"].tensor.tensor_shape.dim[0].size
                )

                if channel_size == 1:
                    max_filter_tensor = []
                    min_filter_tensor = []
                    max_filter_tensor.append((max_filter_node.attr["value"].tensor.float_val)[0])
                    min_filter_tensor.append((min_filter_node.attr["value"].tensor.float_val)[0])
                else:
                    max_filter_tensor = tensor_util.MakeNdarray(max_filter_node.attr["value"].tensor)
                    min_filter_tensor = tensor_util.MakeNdarray(min_filter_node.attr["value"].tensor)

                offset_value = 6
                if (
                    another_conv_node.op == "_FusedQuantizedConv2D"
                    and "fused_ops" in another_conv_node.attr
                    and str(another_conv_node.attr["fused_ops"].list.s) in self.offset_map_new_api
                ):
                    offset_value = self.offset_map_new_api[str(another_conv_node.attr["fused_ops"].list.s)]
                else:
                    offset_value = self.offset_map[another_conv_node.op]

                min_freezed_output_node = self.node_mapping[another_conv_node.input[offset_value]]
                max_freezed_output_node = self.node_mapping[another_conv_node.input[offset_value + 1]]
                min_input = min_freezed_output_node.attr["value"].tensor.float_val[0]
                max_input = max_freezed_output_node.attr["value"].tensor.float_val[0]
                # To avoid generating int32 bias exception for corner case
                if min_input == 0 and max_input == 0:
                    continue

                bias_tensor = tensor_util.MakeNdarray(bias_node.attr["value"].tensor)

                activation_range = 127.0 if current_node.attr["out_type"].type == dtypes.qint8 else 255.0

                int32_bias = Helper.generate_int32_bias_for_conv(
                    bias_tensor,
                    channel_size,
                    max_input,
                    min_input,
                    max_filter_tensor,
                    min_filter_tensor,
                    activation_range,
                )

                original_conv_node.attr["Tbias"].CopyFrom(attr_value_pb2.AttrValue(type=dtypes.qint32.as_datatype_enum))
                bias_node.attr["dtype"].CopyFrom(attr_value_pb2.AttrValue(type=dtypes.qint32.as_datatype_enum))

                bias_node.attr["value"].CopyFrom(
                    attr_value_pb2.AttrValue(
                        tensor=tensor_util.make_tensor_proto(int32_bias, dtypes.int32, bias_tensor.shape)
                    )
                )
                bias_node.attr["value"].tensor.dtype = dtypes.qint32.as_datatype_enum
                if "Thost_inputs" in original_conv_node.attr:
                    original_conv_node.attr["Thost_inputs"].list.type[2] = original_conv_node.attr["Tbias"].type



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tensorflow model calibration process for Smooth Quantization."""

import copy
import logging
import os
import tempfile
import time
from collections import OrderedDict, UserDict

import numpy as np
import tensorflow as tf
from tensorflow.core.framework import attr_value_pb2, graph_pb2
from tensorflow.python.framework import dtypes, tensor_util
from tensorflow.python.saved_model import load, tag_constants

from neural_compressor.tensorflow.quantization.utils.graph_util import GraphAnalyzer
from neural_compressor.tensorflow.quantization.utils.graph_util import GraphRewriterHelper as Helper
from neural_compressor.tensorflow.quantization.utils.quantize_graph_common import QuantizeGraphHelper
from neural_compressor.tensorflow.quantization.utils.utility import (
    iterator_sess_run,
    parse_saved_model,
    reconstruct_saved_model,
)
from neural_compressor.tensorflow.utils import CaptureOutputToFile, TensorflowLLMModel

logger = logging.getLogger("neural_compressor")
debug = bool(logger.level == logging.DEBUG)


class SmoothQuantCalibration:
    """A class for performing smooth quantization calibration on a Tensorflow model.

    Args:
        model (Model): The Tensorflow wrapper model to be calibrated.
        dataloader (DataLoader): The data loader for the calibration dataset.
        iterations (int): The number of iterations to run the calibration process.
        op_types (List[str]): The types of operations to be quantized.
        percentile (float): The percentile of calibration to remove outliers.
    """

    def __init__(self, model, dataloader, iterations, op_types, percentile):
        """Initializes a SmoothQuantCalibration object."""
        self.model = model
        self.dataloader = dataloader
        self.iterations = iterations
        # self.iterations = 3
        self.op_types = op_types
        self.percentile = percentile
        self._sq_input_node_names = []
        self._sq_output_tensor_dict = {}
        self._sq_weight_node_names = {}  # mapping from its weight node name to the concrete output node name

    def _inference_for_calibration(self, model):
        """Run the calibration on the input graph.

        Args:
            model(TensorflowBaseModel): input TensorflowBaseModel
        """
        # ITEX optimization has broken INC calibration process.
        # INC needs turn off ITEX optimization pass in calibration stage.
        # TODO ITEX will provide API to replace setting environment variable.
        os.environ["ITEX_REMAPPER"] = "0"
        sess = model.sess
        iter_op = model.iter_op
        input_tensor = model.input_tensor
        output_tensor = [item + ":0" for item in self._sq_input_node_names]
        # TF table initialization: https://github.com/tensorflow/tensorflow/issues/8665
        node_names = [node.name for node in sess.graph.as_graph_def().node]
        if "init_all_tables" in node_names:  # pragma: no cover
            init_table_op = sess.graph.get_operation_by_name("init_all_tables")
            sess.run(init_table_op)

        logger.info("Start sampling on calibration dataset for Smooth Quantization.")
        if hasattr(self.dataloader, "__len__") and len(self.dataloader) == 0:  # pragma: no cover
            feed_dict = {}
            for output_idx, output in enumerate(
                sess.run(output_tensor, feed_dict)
                if iter_op == []
                else iterator_sess_run(sess, iter_op, feed_dict, output_tensor, self.iterations)
            ):
                self._sq_output_tensor_dict.setdefault(self._sq_input_node_names[output_idx], []).append(output)
        for idx, (inputs, labels) in enumerate(self.dataloader):
            if len(input_tensor) == 1:
                feed_dict = {}
                if (
                    isinstance(inputs, dict) or isinstance(inputs, OrderedDict) or isinstance(inputs, UserDict)
                ):  # pragma: no cover
                    for name in inputs:
                        for tensor in input_tensor:
                            pos = tensor.name.rfind(":")
                            t_name = tensor.name if pos < 0 else tensor.name[:pos]
                            if name == t_name:
                                feed_dict[tensor] = inputs[name]
                                break
                else:
                    feed_dict = {input_tensor[0]: inputs}  # get raw tensor using index [0]
            else:  # pragma: no cover
                assert len(input_tensor) == len(inputs), "inputs len must equal with input_tensor"
                feed_dict = {}
                if isinstance(inputs, dict) or isinstance(inputs, OrderedDict) or isinstance(inputs, UserDict):
                    for name in inputs:
                        for tensor in input_tensor:
                            pos = tensor.name.rfind(":")
                            t_name = tensor.name if pos < 0 else tensor.name[:pos]
                            if name in [tensor.name, t_name]:
                                feed_dict[tensor] = inputs[name]
                                break
                else:
                    # sometimes the input_tensor is not the same order with inputs
                    # we should check and pair them
                    def check_shape(tensor, data):
                        # scalar or 1 dim default True
                        if (
                            tensor.shape is None
                            or tensor.shape.dims is None
                            or len(tensor.shape.dims) == 1
                            or not hasattr(data, "shape")
                        ):
                            return True
                        tensor_shape = tuple(tensor.shape)
                        data_shape = tuple(data.shape)
                        for tensor_dim, data_dim in zip(tensor_shape, data_shape):
                            if tensor_dim is not None and tensor_dim != data_dim:
                                return False
                        return True

                    disorder_tensors = []
                    disorder_inputs = []
                    for idx, sort_tensor in enumerate(input_tensor):
                        sort_input = inputs[idx]
                        if check_shape(sort_tensor, sort_input):
                            feed_dict.update({sort_tensor: sort_input})
                        else:
                            disorder_tensors.append(sort_tensor)
                            disorder_inputs.append(sort_input)
                    for i, dis_tensor in enumerate(disorder_tensors):
                        for j, dis_input in enumerate(disorder_inputs):
                            if check_shape(dis_tensor, dis_input):
                                feed_dict.update({dis_tensor: dis_input})
                                break
            for output_idx, output in enumerate(
                sess.run(output_tensor, feed_dict)
                if iter_op == []
                else iterator_sess_run(sess, iter_op, feed_dict, output_tensor, self.iterations)
            ):
                self._sq_output_tensor_dict.setdefault(self._sq_input_node_names[output_idx], []).append(output)
            if idx + 1 == self.iterations:
                break
        os.environ["ITEX_REMAPPER"] = "1"

    def _generate_calibration_data(self):
        """Generate the calibration data."""
        sorted_graph = QuantizeGraphHelper().get_sorted_graph(
            self.model.graph_def, self.model.input_node_names, self.model.output_node_names
        )

        for node in sorted_graph.node:
            if node.op not in self.op_types:
                continue
            # Fix retval already been set issue
            if "while" in node.input[0]:  # pragma: no cover
                continue
            self._sq_input_node_names.append(node.input[0])
            self._sq_weight_node_names[node.input[1]] = node.name

        self._inference_for_calibration(self.model)

    def _get_maxval_per_channel(self, tensor_data, percentile):
        """Get the max values per input channel.

        Args:
            tensor_data: The input tensors
            percentile: The percentile of calibration to remove outliers

        Returns:
            The max values per input channel
        """
        permute_datas = []
        for data in tensor_data:  # iteration_num * (N, H, W, C)
            if len(data.shape) == 3:  # pragma: no cover
                # TODO  matmul batchsize*seq*inchannel
                tensor = np.abs(np.reshape(data, (-1, data.shape[-1])))
                permute_datas.append(tensor)
            elif len(data.shape) == 4:  # already NHWC
                # tensor = np.transpose(data, [0, 3, 1, 2])
                tensor = data
                tensor = np.abs(np.reshape(tensor, (-1, tensor.shape[-1])))
                permute_datas.append(tensor)
            elif len(data.shape) == 2:  # (?, ic)
                permute_datas.append(np.abs(data))
            else:  # pragma: no cover
                assert False, "not supported"
        permute_datas = np.concatenate(permute_datas, axis=0)
        permute_datas = permute_datas.reshape(-1, permute_datas.shape[-1])
        # try:
        #     np.percentile(permute_datas, percentile, axis=0)
        # except FloatingPointError:
        #     indexes = [i for i,e in enumerate(np.percentile(permute_datas, percentile, axis=0)) if np.isnan(e)][0]
        #     np.seterr(all='warning')
        max_per_channels = np.percentile(permute_datas, percentile, axis=0)
        # max_per_channels = np.max(permute_datas, axis=0)
        max_per_channels = max_per_channels.astype(np.single)
        return max_per_channels

    def __call__(self):
        """Generates calibration data and calculate the maximum values per channel.

        Returns:
            max_vals_per_channel (dict): A dictionary containing the maximum values per channel.
            sq_weight_node_names (dict): A dictionary mapping from weight names to target node names.
        """
        self._generate_calibration_data()
        max_vals_per_channel = {}
        for key in self._sq_output_tensor_dict.keys():
            max_val_per_channel = self._get_maxval_per_channel(
                self._sq_output_tensor_dict[key], percentile=self.percentile
            )
            max_vals_per_channel[key] = max_val_per_channel
        return max_vals_per_channel, self._sq_weight_node_names


class SmoothQuantCalibrationLLM(SmoothQuantCalibration):
    """A class for performing smooth quantization calibration on a Tensorflow LLM model.

    Args:
        model (str): A path to the original Tensorflow model.
        iterations (int): The number of iterations to run the calibration process.
        op_types (List[str]): The types of operations to be quantized.
        percentile (float): The percentile of calibration to remove outliers.
        eval_func (function):  The function to inference the model.
        temp_path (str): The temporary path to store median model.
        weight_name_mapping (): A function that convert weight tensor name in autotrackable to node name in graph_def
    """

    def __init__(self, model_path, dataloader, iterations, op_types, percentile, temp_path, weight_name_mapping):
        """Initializes a SmoothQuantCalibrationLLM object."""
        self.func = None
        self.graph_def = None
        self.frozen_func = None
        self._saved_model = None
        self.model = model_path
        self.dataloader = dataloader
        self.iterations = iterations
        self.op_types = op_types
        self.percentile = percentile
        self.temp_path = temp_path
        self.weight_name_mapping = weight_name_mapping
        self.print_node_list = []
        self._sq_input_node_names = []
        self._sq_target_node_names = {}
        self._sq_output_tensor_dict = {}
        self._sq_weight_tensor_dict = {}

    def _parse_calibration_logs(self, tmp_dump_file):
        """Parse calibration logs for llm saved_model."""
        valid_data = []
        with open(tmp_dump_file) as file:
            for i in file.readlines():
                if i.startswith(";"):
                    valid_data.append(i.strip())

        for activation in valid_data:
            [key, value] = activation.rsplit(":")
            activation_name = key[1:-9]
            import json

            value = value.replace(" ", ",")
            value = value.replace("][", "],[")
            data = json.loads(value)
            if activation_name not in self._sq_output_tensor_dict:
                self._sq_output_tensor_dict[activation_name] = [np.array(data)]
            else:
                self._sq_output_tensor_dict[activation_name].append(np.array(data))

    def _insert_print_for_activation(self, graph_def):
        """Insert print node in the graph to do the calibration for llm saved_model."""
        cur_graph = GraphAnalyzer()
        cur_graph.graph = graph_def

        graph_info = cur_graph.parse_graph()
        for cur_list in self.print_node_list:
            pre_node_name = cur_list[0]
            post_node_name = cur_list[-1]
            insert_node_pairs = []
            top_node = graph_info[pre_node_name].node
            if top_node.op == "ConcatV2":
                for i in range(top_node.attr["N"].i):
                    insert_node_pairs.append([top_node.input[i], post_node_name])
            elif top_node.op in ("BatchMatMul", "BatchMatMulV2"):
                insert_node_pairs.append([top_node.input[0], post_node_name])
                if graph_info[top_node.input[1]].node.op != "Const":
                    insert_node_pairs.append([top_node.input[1], post_node_name])
            elif top_node.op in ("Conv2DBackpropInput", "Conv3DBackpropInputV2"):
                insert_node_pairs.append([top_node.input[2], post_node_name])
            else:
                refresh_pre_node_name = graph_info[pre_node_name].node.input[0]
                # Check the Conv2D could be fused with previous Pad or not.
                # If so, we need to update the pre-node name correspondingly.
                refresh_pre_node = graph_info[Helper.node_name_from_input(refresh_pre_node_name)].node
                if refresh_pre_node.op == "Pad" and top_node.op in ("Conv2D", "Conv3D"):
                    insert_node_pairs.append([refresh_pre_node_name, post_node_name])
                    refresh_pre_node_name = refresh_pre_node.input[0]

                insert_node_pairs.append([refresh_pre_node_name, post_node_name])

            output_names = []
            for node_pair_names in insert_node_pairs:
                for index, each_node_name in enumerate(node_pair_names):
                    name_with_sig = each_node_name
                    node_name_prefix = name_with_sig.replace(":", "__port__").replace("^", "__hat__")
                    print_node = Helper.create_node(
                        "Print",
                        node_name_prefix + "_print__{}".format(index),
                        [each_node_name + ":0", each_node_name + ":0"],
                    )

                    if index == 0:
                        msg = ";{}__print__:".format(each_node_name)
                        # workaround for swish_f32, attribute T is not in the op definition
                        if "swish_f32" in graph_info[pre_node_name].node.name:
                            src_dt = attr_value_pb2.AttrValue(type=dtypes.float32.as_datatype_enum)
                        else:
                            src_dt = graph_info[pre_node_name].node.attr["T"]
                    else:
                        break

                    print_node.attr["T"].CopyFrom(src_dt)

                    print_node.attr["message"].s = msg.encode()
                    print_node.attr["first_n"].i = -1
                    print_node.attr["summarize"].i = 102400000

                    attr_u = [dtypes.as_dtype(src_dt.type).as_datatype_enum]
                    print_node.attr["U"].list.CopyFrom(attr_value_pb2.AttrValue.ListValue(type=attr_u))
                    post_node_names = graph_info[Helper.node_name_from_input(each_node_name)].outputs
                    if post_node_names:
                        for post_node_name in post_node_names:
                            post_node = graph_info[post_node_name].node
                            if each_node_name not in post_node.input:
                                continue
                            if (
                                post_node.op == "FusedBatchNormV3"
                                and "_print_identity"
                                not in graph_info[Helper.node_name_from_input(post_node.name)].node.input[0]
                            ):
                                identity_node = Helper.create_node(
                                    "Identity",
                                    post_node.name + "_print_identity",
                                    [graph_info[Helper.node_name_from_input(post_node.name)].node.input[0]],
                                )
                                identity_node.attr["T"].CopyFrom(src_dt)
                                cur_graph.add_node(
                                    identity_node,
                                    graph_info[Helper.node_name_from_input(post_node.name)].node.input[0],
                                    [post_node.name],
                                )
                                identity_node.input.append("^" + print_node.name)
                            else:
                                post_node.input.append("^" + print_node.name)

                        cur_graph.add_node(print_node, each_node_name, [])
                    else:
                        identity_node1 = Helper.create_node(
                            "Identity", print_node.name + "_identity", [print_node.name]
                        )
                        identity_node1.attr["T"].CopyFrom(src_dt)
                        cur_graph.add_node(print_node, each_node_name, [identity_node1.name])
                        cur_graph.add_node(identity_node1, print_node.name, [])
                        output_names.append(identity_node1.name)

        return cur_graph.dump_graph()

    def evaluate(self, model):
        """Evaluate function that inference the model to apply calibration.

        Args:
            model (tf.python.trackable.autotrackable): The model to be evaluated.
            The object is usually gotten by using tf.saved_model.load(model_dir) API.

        Returns:
            accuracy (float): The accuracy result.
        """
        input_tensor_names = model.input_tensor_names
        auto_trackable = model.model
        infer = auto_trackable.signatures["serving_default"]
        for idx, (inputs, _) in enumerate(self.dataloader):
            feed_dict = {}
            if len(input_tensor_names) == 1:
                feed_dict[input_tensor_names[0]] = inputs
            else:
                assert len(input_tensor_names) == len(inputs), "inputs len must equal with input_tensor"
                for i, input_tensor_name in enumerate(input_tensor_names):
                    feed_dict[input_tensor_name] = inputs[i]

            _ = infer(**feed_dict)

            if idx >= self.iterations:
                break

    def _inference(self, sampling_graph_def):
        """Inference the model to apply calibration.

        Args:
            sampling_graph_def: The temporary graph_def for inference.
        """
        logger.info("Start sampling on calibration dataset for Smooth Quantization.")
        # reconstruct graph_def that inserted print node to saved_model
        reconstruct_saved_model(sampling_graph_def, self.func, self.frozen_func, self._saved_model, self.temp_path)
        model = TensorflowLLMModel(self.temp_path)

        input_tensor_names = model.input_tensor_names
        auto_trackable = model.model
        infer = auto_trackable.signatures["serving_default"]
        for idx, (inputs, _) in enumerate(self.dataloader):
            feed_dict = {}
            if len(input_tensor_names) == 1:
                feed_dict[input_tensor_names[0]] = inputs
            else:
                assert len(input_tensor_names) == len(inputs), "inputs len must equal with input_tensor"
                for i, input_tensor_name in enumerate(input_tensor_names):
                    feed_dict[input_tensor_name] = inputs[i]

            _ = infer(**feed_dict)

            if idx >= self.iterations:
                break

    def _inference_for_calibration(self, model):
        """Run the calibration on the input graph."""
        sampling_graph_def = self._insert_print_for_activation(model)
        tmp_dump_file = tempfile.mkstemp(suffix=".log")[1]
        with CaptureOutputToFile(tmp_dump_file):
            self._inference(sampling_graph_def)
        self._parse_calibration_logs(tmp_dump_file)
        del sampling_graph_def

    def _get_weight_tensors(self):
        model = load.load(self.model, [tag_constants.SERVING])
        for weight_tensor in model.variables:
            parsed_name = self.weight_name_mapping(weight_tensor.name)
            if parsed_name in self._sq_target_node_names:
                self._sq_weight_tensor_dict[parsed_name] = weight_tensor.numpy()

        assert len(self._sq_weight_tensor_dict) == len(
            self._sq_target_node_names
        ), "Failed to get weights for some nodes, please check variables"

    def _generate_calibration_data(self, input_node_names, output_node_names):
        """Generate the calibration data."""
        sorted_graph = QuantizeGraphHelper().get_sorted_graph(
            self.graph_def,
            input_node_names,
            output_node_names,
        )

        for node in sorted_graph.node:
            if node.op not in self.op_types:
                continue
            # Fix retval already been set issue
            if "while" in node.input[0]:  # pragma: no cover
                continue
            self._sq_input_node_names.append(node.input[0])
            self.print_node_list.append([node.name])
            self._sq_target_node_names[node.input[1]] = node.name
        self._get_weight_tensors()
        sampling_graph_def = copy.deepcopy(self.graph_def)
        self._inference_for_calibration(sampling_graph_def)

    def __call__(self, input_node_names, output_node_names):
        """Generates calibration data and calculate the maximum values per channel.

        Args:
            input_node_names: (list): A list of names for input nodes.
            output_node_names: (list): A list of names for output nodes.

        Returns:
            max_vals_per_channel (dict): A dictionary containing the maximum values per channel.
            sq_target_node_names (dict): A dictionary mapping from weight names to target node names.
            sq_weight_tensor_dict (dict): A dictionary containing tensor of weights.
        """
        self.graph_def, self._saved_model, self.func, self.frozen_func, _, _ = parse_saved_model(self.model)
        self._generate_calibration_data(input_node_names, output_node_names)
        max_vals_per_channel = {}
        for activation_name, output_tensor in self._sq_output_tensor_dict.items():
            max_val_per_channel = self._get_maxval_per_channel(output_tensor, percentile=self.percentile)
            max_vals_per_channel[activation_name] = max_val_per_channel
        return max_vals_per_channel, self._sq_target_node_names, self._sq_weight_tensor_dict, self.graph_def



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Quantize Conv2D/DepthwiseConv2dNative."""

import numpy as np
import tensorflow as tf
from tensorflow.core.framework import graph_pb2, node_def_pb2
from tensorflow.python.framework import dtypes, tensor_util

from neural_compressor.tensorflow.quantization.utils.quantize_graph_common import QuantizeGraphHelper as helper
from neural_compressor.tensorflow.utils import version1_gte_version2

from .quantize_graph_base import QuantizeNodeBase


class FuseNodeStartWithConv2d(QuantizeNodeBase):
    """Quantize Conv2D/DepthwiseConv2dNative to int8 op."""

    def __init__(self, **kwargs):
        """Initialization."""
        super().__init__(**kwargs)
        self.sorted_patterns = sorted(self.patterns, key=lambda i: len(i), reverse=True)
        self.exclude_conv_name = []

        self.fusion_mapping = {
            "Conv2DBiasAdd": self.apply_conv_biasadd_fusion,
            "Conv2DBiasAddAddNRelu": self.apply_conv_biasadd_addn_relu_fusion,
            "Conv2DBiasAddAddNRelu6": self.apply_conv_biasadd_addn_relu_fusion,
            "Conv2DBiasAddAddV2Relu": self.apply_conv_biasadd_addn_relu_fusion,
            "Conv2DBiasAddAddV2Relu6": self.apply_conv_biasadd_addn_relu_fusion,
            "Conv2DBiasAddAddRelu": self.apply_conv_biasadd_addn_relu_fusion,
            "Conv2DBiasAddRelu6": self.apply_conv_biasadd_relu_fusion,
            "Conv2DBiasAddRelu": self.apply_conv_biasadd_relu_fusion,
            "Conv2DBiasAddLeakyRelu": self.apply_conv_biasadd_relu_fusion,
            "Conv2DBiasAddLeakyReluAddV2": self.apply_conv_biasadd_addn_relu_fusion,
            "Conv2DAddRelu6": self.apply_conv_biasadd_relu_fusion,
            "Conv2DAddRelu": self.apply_conv_biasadd_relu_fusion,
            "DepthwiseConv2dNativeAddRelu6": self.apply_conv_biasadd_relu_fusion,
            "DepthwiseConv2dNativeBiasAddRelu": self.apply_conv_biasadd_relu_fusion,
            "DepthwiseConv2dNativeBiasAdd": self.apply_conv_biasadd_fusion,
            "DepthwiseConv2dNativeBiasAddRelu6": self.apply_conv_biasadd_relu_fusion,
            "Conv2D": self.apply_conv_single_fusion,
            "DepthwiseConv2dNative": self.apply_conv_single_fusion,
        }

    def apply_conv_single_fusion(self, match_node_name):
        """Apply Conv2D single fusion."""
        skip_node_name = match_node_name[1:]
        matched_node = self.node_name_mapping[match_node_name[0]]
        _, normal_inputs = self._get_node_input(matched_node.node.name)
        weight_name = normal_inputs[1]
        # TODO this is workaround as the tf 2.1 doesn't support depthwise/conv s8
        # feature.
        if self.enable_s8 and not self._find_relu_node(matched_node.node):
            self.exclude_conv_name.append(match_node_name[0])
            self.output_graph = self.input_graph
            return

        q_weights_name, q_weights_min_name, q_weights_max_name = self._intel_cpu_quantize_weight_eightbit(
            matched_node.node.op, self.node_name_mapping[weight_name].node, self.per_channel
        )

        all_input_names = self._add_eightbit_prologue_nodes(matched_node.node.name)
        all_input_names = all_input_names[:1] + [q_weights_name] + all_input_names[1:]
        all_input_names.append(q_weights_min_name)
        all_input_names.append(q_weights_max_name)
        skip_node_name.append(weight_name)

        for _, node in enumerate(self.input_graph.node):
            if node.name in skip_node_name:
                self.logger.debug("Skip node {}.".format(node.name))
            elif node.name == match_node_name[0]:
                postfix = "_eightbit_quantized_depthwise_conv"
                if node.op == "Conv2D":
                    postfix = "_eightbit_quantized_conv"
                quantized_node_name = node.name + postfix
                if node.op == "Conv2D":
                    quantized_conv_node = helper.create_node(
                        "QuantizedConv2DPerChannel" if self.per_channel else "QuantizedConv2D",
                        quantized_node_name,
                        all_input_names,
                    )

                elif node.op == "DepthwiseConv2dNative":
                    quantized_conv_node = helper.create_node(
                        "QuantizedDepthwiseConv2D", quantized_node_name, all_input_names
                    )

                helper.copy_attr(quantized_conv_node, "strides", node.attr["strides"])
                helper.copy_attr(quantized_conv_node, "padding", node.attr["padding"])
                if "padding_list" in node.attr:
                    helper.copy_attr(quantized_conv_node, "padding_list", node.attr["padding_list"])
                helper.copy_attr(quantized_conv_node, "dilations", node.attr["dilations"])
                input_data_type = dtypes.quint8 if self._find_relu_node(node) else dtypes.qint8
                helper.set_attr_dtype(quantized_conv_node, "Tinput", input_data_type)
                helper.set_attr_dtype(quantized_conv_node, "Tfilter", dtypes.qint8)
                helper.set_attr_dtype(quantized_conv_node, "out_type", dtypes.qint32)
                self.add_output_graph_node(quantized_conv_node)
                quantize_down_name = self._add_quantize_down_nodes(node, quantized_node_name, dtypes.qint8)
                self._intel_cpu_add_dequantize_result_node(quantize_down_name, node.name, dtypes.qint8)
            else:
                new_node = node_def_pb2.NodeDef()
                new_node.CopyFrom(node)
                self.add_output_graph_node(new_node)

    def apply_conv_biasadd_relu_fusion(self, match_node_name):
        """Fuse the conv/biasadd/relu pattern."""
        skip_node_name = match_node_name[1:]
        matched_node = self.node_name_mapping[match_node_name[0]]
        control_inputs, normal_inputs = self._get_node_input(matched_node.node.name)
        weight_name = normal_inputs[1]

        q_weights_name, q_weights_min_name, q_weights_max_name = self._intel_cpu_quantize_weight_eightbit(
            matched_node.node.op, self.node_name_mapping[weight_name].node, self.per_channel
        )

        all_input_names = self._add_eightbit_prologue_nodes(matched_node.node.name)
        all_input_names = all_input_names[:1] + [q_weights_name] + all_input_names[1:]
        all_input_names.append(q_weights_min_name)
        all_input_names.append(q_weights_max_name)
        skip_node_name.append(weight_name)

        for _, node in enumerate(self.input_graph.node):
            if node.name in skip_node_name:
                self.logger.debug("Skip node {}.".format(node.name))
            elif node.name == match_node_name[0]:
                postfix = "_eightbit_quantized_depthwise_conv"
                if node.op == "Conv2D":
                    postfix = "_eightbit_quantized_conv"
                quantized_node_name = node.name + postfix
                bias_node_name = self.node_name_mapping[match_node_name[1]].node.input[1]
                relu_node_name = match_node_name[2]
                is_relu6 = self.node_name_mapping[relu_node_name].node.op == "Relu6"
                quantized_node_input_names = (
                    all_input_names[:2] + [bias_node_name] + all_input_names[2:] + control_inputs
                )
                is_leakyrelu = self.node_name_mapping[relu_node_name].node.op == "LeakyRelu"
                quantized_conv_node_op = "QuantizedDepthwiseConv2DWithBiasAndRelu"
                if node.op == "Conv2D" or is_leakyrelu:
                    quantized_conv_node_op = "QuantizedConv2DWithBiasAndRelu"
                quantized_conv_node = helper.create_node(
                    quantized_conv_node_op, quantized_node_name, quantized_node_input_names
                )
                helper.copy_attr(quantized_conv_node, "strides", node.attr["strides"])
                helper.copy_attr(quantized_conv_node, "padding", node.attr["padding"])
                if "alpha" in self.node_name_mapping[relu_node_name].node.attr:
                    helper.copy_attr(
                        quantized_conv_node, "alpha", self.node_name_mapping[relu_node_name].node.attr["alpha"]
                    )
                if "padding_list" in node.attr:
                    helper.copy_attr(quantized_conv_node, "padding_list", node.attr["padding_list"])
                helper.copy_attr(quantized_conv_node, "dilations", node.attr["dilations"])
                input_data_type = dtypes.quint8 if self._find_relu_node(node) else dtypes.qint8
                helper.set_attr_dtype(quantized_conv_node, "Tinput", input_data_type)
                helper.set_attr_dtype(quantized_conv_node, "Tfilter", dtypes.qint8)
                helper.set_attr_dtype(quantized_conv_node, "out_type", dtypes.qint32)
                self.add_output_graph_node(quantized_conv_node)
                if not is_leakyrelu:
                    quantize_down_name = self._add_quantize_down_nodes(
                        node, quantized_node_name, dtypes.quint8, is_relu6
                    )
                    self._intel_cpu_add_dequantize_result_node(quantize_down_name, relu_node_name)
                else:
                    quantize_down_name = self._add_quantize_down_nodes(node, quantized_node_name, dtypes.qint8, False)
                    self._intel_cpu_add_dequantize_result_node(quantize_down_name, relu_node_name, dtype=dtypes.qint8)
            else:
                new_node = node_def_pb2.NodeDef()
                new_node.CopyFrom(node)
                self.add_output_graph_node(new_node)

    def apply_conv_biasadd_fusion(self, match_node_name):
        """Apply Conv BiasAdd fusion."""
        skip_node_name = match_node_name[1:]
        matched_node = self.node_name_mapping[match_node_name[0]]
        control_inputs, normal_inputs = self._get_node_input(matched_node.node.name)
        weight_name = normal_inputs[1]

        q_weights_name, q_weights_min_name, q_weights_max_name = self._intel_cpu_quantize_weight_eightbit(
            matched_node.node.op, self.node_name_mapping[weight_name].node, self.per_channel
        )

        all_input_names = self._add_eightbit_prologue_nodes(matched_node.node.name)
        all_input_names = all_input_names[:1] + [q_weights_name] + all_input_names[1:]
        all_input_names.append(q_weights_min_name)
        all_input_names.append(q_weights_max_name)
        skip_node_name.append(weight_name)

        for _, node in enumerate(self.input_graph.node):
            if node.name in skip_node_name:
                pass
            elif node.name == match_node_name[0]:
                self.logger.debug("Matched node {} with input {}.".format(node.name, node.input))

                quantized_node_name = node.name + "_eightbit_quantized_conv"
                if node.op == "DepthwiseConv2dNative":
                    quantized_node_name = node.name + "_eightbit_quantized_depthwise_conv"

                bias_node_name = self.node_name_mapping[match_node_name[1]].node.input[1]
                quantized_node_input_names = (
                    all_input_names[:2] + [bias_node_name] + all_input_names[2:] + control_inputs
                )

                quantized_conv_node = helper.create_node(
                    "QuantizedConv2DWithBias" if node.op == "Conv2D" else "QuantizedDepthwiseConv2DWithBias",
                    quantized_node_name,
                    quantized_node_input_names,
                )

                helper.copy_attr(quantized_conv_node, "strides", node.attr["strides"])
                helper.copy_attr(quantized_conv_node, "padding", node.attr["padding"])
                if "padding_list" in node.attr:
                    helper.copy_attr(quantized_conv_node, "padding_list", node.attr["padding_list"])
                helper.copy_attr(quantized_conv_node, "dilations", node.attr["dilations"])

                input_data_type = dtypes.quint8 if self._find_relu_node(node) else dtypes.qint8

                helper.set_attr_dtype(quantized_conv_node, "Tinput", input_data_type)
                helper.set_attr_dtype(quantized_conv_node, "Tfilter", dtypes.qint8)
                helper.set_attr_dtype(quantized_conv_node, "out_type", dtypes.qint32)
                self.add_output_graph_node(quantized_conv_node)
                requantize_type = dtypes.qint8

                quantize_down_name = self._add_quantize_down_nodes(node, quantized_node_name, requantize_type, False)
                self._intel_cpu_add_dequantize_result_node(quantize_down_name, match_node_name[1], requantize_type)
            else:
                new_node = node_def_pb2.NodeDef()
                new_node.CopyFrom(node)
                self.add_output_graph_node(new_node)

    def apply_conv_biasadd_addn_relu_fusion(self, match_node_name):
        """Apply Conv BiasAdd AddN Relu fusion."""
        skip_node_name = match_node_name[1:]
        matched_node = self.node_name_mapping[match_node_name[0]]
        control_inputs, normal_inputs = self._get_node_input(matched_node.node.name)
        weight_name = normal_inputs[1]

        third_node = self.node_name_mapping[match_node_name[2]].node
        forth_node = self.node_name_mapping[match_node_name[3]].node
        if third_node.op != "LeakyRelu" and not self._find_relu_node(matched_node.node):
            return self.apply_conv_biasadd_fusion(match_node_name[:2])

        is_leakyrelu_add_fusion = third_node.op == "LeakyRelu" and forth_node.op.find("Add") != -1

        q_weights_name, q_weights_min_name, q_weights_max_name = self._intel_cpu_quantize_weight_eightbit(
            matched_node.node.op, self.node_name_mapping[weight_name].node, self.per_channel
        )

        all_input_names = self._add_eightbit_prologue_nodes(matched_node.node.name)
        all_input_names = all_input_names[:1] + [q_weights_name] + all_input_names[1:]
        all_input_names.append(q_weights_min_name)
        all_input_names.append(q_weights_max_name)
        skip_node_name.append(weight_name)

        for _, node in enumerate(self.input_graph.node):
            if node.name in skip_node_name:
                self.logger.debug("skip node {}".format(node.name))
            elif node.name == match_node_name[0]:
                self.logger.debug("Matched node {} with input {}.".format(node.name, node.input))

                quantized_node_name = node.name + "_eightbit_quantized_conv"

                bias_node_name = self.node_name_mapping[match_node_name[1]].node.input[1]
                leaky_offset = 1 if is_leakyrelu_add_fusion else 0
                if is_leakyrelu_add_fusion:
                    relu_node_name = match_node_name[2]
                else:
                    relu_node_name = match_node_name[3]
                is_relu6 = self.node_name_mapping[relu_node_name].node.op == "Relu6"

                sum_index = (
                    1
                    if match_node_name[1 + leaky_offset]
                    == self.node_name_mapping[match_node_name[2 + leaky_offset]].node.input[0]
                    else 0
                )
                quantized_node_input_names = (
                    all_input_names[:2]
                    + [bias_node_name]
                    + all_input_names[2:]
                    + [self.node_name_mapping[match_node_name[2 + leaky_offset]].node.input[sum_index]]
                    + control_inputs
                )

                node_op = (
                    "QuantizedConv2DWithBiasReluAndSum"
                    if is_leakyrelu_add_fusion
                    else "QuantizedConv2DWithBiasSumAndRelu"
                )

                quantized_conv_node = helper.create_node(node_op, quantized_node_name, quantized_node_input_names)
                helper.copy_attr(quantized_conv_node, "strides", node.attr["strides"])
                helper.copy_attr(quantized_conv_node, "padding", node.attr["padding"])
                if "padding_list" in node.attr:
                    helper.copy_attr(quantized_conv_node, "padding_list", node.attr["padding_list"])
                helper.copy_attr(quantized_conv_node, "dilations", node.attr["dilations"])
                input_data_type = dtypes.quint8 if self._find_relu_node(node) else dtypes.qint8
                helper.set_attr_dtype(quantized_conv_node, "Tinput", input_data_type)
                helper.set_attr_dtype(quantized_conv_node, "Tfilter", dtypes.qint8)
                helper.set_attr_dtype(quantized_conv_node, "out_type", dtypes.qint32)
                if "alpha" in self.node_name_mapping[relu_node_name].node.attr:
                    helper.copy_attr(
                        quantized_conv_node, "alpha", self.node_name_mapping[relu_node_name].node.attr["alpha"]
                    )
                self.add_output_graph_node(quantized_conv_node)

                if is_leakyrelu_add_fusion:
                    quantize_down_name = self._add_quantize_down_nodes(node, quantized_node_name, dtypes.qint8, False)
                    self._intel_cpu_add_dequantize_result_node(
                        quantize_down_name, match_node_name[3], dtype=dtypes.qint8
                    )
                else:
                    dtype = dtypes.quint8
                    if [
                        i
                        for i in self.node_name_mapping[relu_node_name].output
                        if "FusedBatchNorm" in self.node_name_mapping[i].node.op and i in self.op_wise_config_name_list
                    ]:
                        dtype = dtypes.qint8
                    quantize_down_name = self._add_quantize_down_nodes(node, quantized_node_name, dtype, is_relu6)
                    self._intel_cpu_add_dequantize_result_node(quantize_down_name, relu_node_name, dtype)
            else:
                new_node = node_def_pb2.NodeDef()
                new_node.CopyFrom(node)
                self.add_output_graph_node(new_node)

    def get_longest_fuse(self):
        """Get the longest fusion pattern."""
        self._get_op_list()

        matched_rule, matched_node_name = self._is_match(self.sorted_patterns)
        return matched_rule, matched_node_name

    def apply_the_transform(self):
        """Quantize Conv2D/DepthwiseConv2dNative and apply the fusion pattern."""
        self._get_op_list()
        matched_rule, matched_node_name = self._is_match(self.sorted_patterns)
        if matched_node_name:
            self.output_graph = graph_pb2.GraphDef()
            fusion_name = "".join(matched_rule)
            if fusion_name in self.fusion_mapping:
                if fusion_name.find("Conv2DAddRelu") != -1:
                    for input_name in self.node_name_mapping[matched_node_name[1]].node.input:
                        input_node_name = helper.node_name_from_input(input_name)
                        if input_node_name != matched_node_name[0]:
                            add_const_input_node = self.node_name_mapping[input_node_name].node
                            add_node_content = tensor_util.MakeNdarray(add_const_input_node.attr["value"].tensor)
                            if add_node_content.ndim != 1:
                                fusion_name = "Conv2D"
                                matched_node_name = matched_node_name[:1]
                self.fusion_mapping[fusion_name](matched_node_name)
            else:  # pragma: no cover
                self.logger.info("Unknown fusion pattern {}.".format(fusion_name))
                if self.remove_redundant_quant_flag:
                    self.input_graph = self.remove_redundant_quantization(self.input_graph)
                return self.input_graph, []

            self.input_graph = self.output_graph
            self._reset_output_node_maps()
            if self.remove_redundant_quant_flag:
                self.output_graph = self.remove_redundant_quantization(self.output_graph)

            return self.output_graph, matched_node_name, self.exclude_conv_name

        return self.input_graph, [], []



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Inject dummy BiasAdd Graph Rewriter."""

import tensorflow as tf
from tensorflow.python.framework import dtypes

from neural_compressor.tensorflow.quantization.utils.graph_rewriter.graph_base import GraphRewriterBase
from neural_compressor.tensorflow.quantization.utils.graph_util import GraphAnalyzer
from neural_compressor.tensorflow.quantization.utils.graph_util import GraphRewriterHelper as Helper
from neural_compressor.tensorflow.utils import dump_elapsed_time


class InjectDummyBiasAddOptimizer(GraphRewriterBase):
    """Inject dummy BiasAdd for MatMul, Conv2D for pattern fusion."""

    def __init__(self, model, outputs):
        """Initialization."""
        super().__init__(model)
        self.outputs = outputs

    @dump_elapsed_time("Pass InjectDummyBiasAddOptimizer")
    def do_transformation(self):
        """Inject dummy BiasAdd if MatMul, Conv2D missing the valid add ops behind them."""
        g = GraphAnalyzer()
        g.graph = self.model
        graph_info = g.parse_graph()
        g.get_frame_info()
        valid_ops = ("BiasAdd", "Add", "AddV2", "AddN")
        target_nodes = g.query_fusion_pattern_nodes(
            [
                ["MatMul", "Conv2D"],
            ]
        )
        for i in target_nodes:
            # only apply this pass for tensorflow old quantization API, pre_optimize does this check
            # use conv+dummy_biasadd+relu because TF do not support conv+relu now.
            if i[0] in self.outputs:
                continue
            next_node_names = graph_info[i[0]].outputs
            if (
                next_node_names
                and len(next_node_names) == 1
                and graph_info[Helper.node_name_from_input(next_node_names[0])].node.op in valid_ops
            ):
                continue
            bias_node_name = i[0] + "_dummy_biasadd"
            bias_const_node_name = i[0] + "_dummy_biasadd_const"
            matmul_a_node_name = Helper.node_name_from_input(graph_info[i[0]].node.input[0])
            matmul_a_node = graph_info[matmul_a_node_name].node
            matmul_b_node_name = Helper.node_name_from_input(graph_info[i[0]].node.input[1])
            matmul_b_node = graph_info[matmul_b_node_name].node

            if matmul_a_node.op == "Const" or matmul_b_node.op not in ["Const", "Enter"]:
                continue
            if matmul_b_node.op == "Enter":  # pragma: no cover
                parent_node = graph_info[Helper.node_name_from_input(matmul_b_node.input[0])].node
                if parent_node.op != "Const":
                    continue
                else:
                    matmul_b_node = parent_node
                    matmul_b_node_name = matmul_b_node.name

            if graph_info[i[0]].node.op == "MatMul":
                t_b_index = 0 if graph_info[i[0]].node.attr["transpose_b"].b else 1
            elif graph_info[i[0]].node.op == "Conv2D" and graph_info[i[0]].node.attr["data_format"].s == b"NHWC":
                t_b_index = 3
            elif graph_info[i[0]].node.op == "Conv2D" and graph_info[i[0]].node.attr["data_format"].s == b"NCHW":
                t_b_index = 1
            else:
                continue

            bias_add_length = matmul_b_node.attr["value"].tensor.tensor_shape.dim[t_b_index].size

            bias_add_content = [0.0] * bias_add_length

            bias_const_node = Helper.create_constant_node(
                bias_const_node_name, bias_add_content, dtypes.float32, shape=[bias_add_length]
            )

            if i[0] in g.parent_frame_details and g.parent_frame_details[i[0]]:  # pragma: no cover
                bias_const_enter_node = Helper.create_node(
                    "Enter", bias_const_node_name + "_enter", [bias_const_node_name]
                )
                Helper.set_attr_string(
                    bias_const_enter_node, "frame_name", g.parent_frame_details[i[0]].attr["frame_name"].s
                )
                Helper.set_attr_dtype(bias_const_enter_node, "T", dtypes.float32)
                Helper.set_attr_bool(bias_const_enter_node, "is_constant", True)
                Helper.set_attr_int(
                    bias_const_enter_node,
                    "parallel_iterations",
                    g.parent_frame_details[i[0]].attr["parallel_iterations"].i,
                )

            bias_node = Helper.create_node(
                "BiasAdd",
                bias_node_name,
                [
                    i[0],
                    (
                        bias_const_enter_node.name
                        if i[0] in g.parent_frame_details and g.parent_frame_details[i[0]]
                        else bias_const_node_name
                    ),
                ],
            )
            Helper.set_attr_dtype(bias_node, "T", dtypes.float32)
            g.add_node(bias_node, i[0], next_node_names)
            if i[0] in g.parent_frame_details and g.parent_frame_details[i[0]]:  # pragma: no cover
                g.add_node(bias_const_node, None, [bias_const_enter_node.name])
                g.add_node(bias_const_enter_node, bias_const_node_name, [bias_node_name])
            else:
                g.add_node(bias_const_node, None, [bias_node_name])

        return g.dump_graph()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Fuse small ops to LayerNorm Graph Rewriter."""

import re

from tensorflow.core.framework import attr_value_pb2, graph_pb2, node_def_pb2
from tensorflow.python.framework import tensor_util

from neural_compressor.tensorflow.utils import dump_elapsed_time


class FuseLayerNormOptimizer:  # pragma: no cover
    """Remap smaller ops into fused LayerNorm.

    Current fusion is only for the case, when LayerNormalization uses FusedBatcNormV3.
    And further restrict it to only 2D or 3D tensor inputs to keras LayerNormalization api.
    """

    def __init__(self, input_graph_def):
        """Constructor."""
        self.input_graph_def = input_graph_def

    @dump_elapsed_time("Pass FuseLayerNormOptimizer")
    def do_transformation(self):
        """The following pattern will be searched in the graph with additional constraints.

        Here * means any type of op.
        Subgraph:
                *(input)  *  * Const  *  Const                       FusedOp
                     x    |   x  |    |  x        Const              -------
                      x   |    x |    | x  Const   x
                      Reshape  Fill   Fill  x     x         *(input) *(gamma)  *(beta)
                         x      x      x   x     x                x     |      x
                          x    x      x   x     x                  x    |     x
                     F u s e d B a t c h N o r m V 3              _MklLayerNorm
                            x
                             x   *
                              x x
                           Reshape
                               x   *(gamma)
                                x x
                                Mul
                        *(beta) x
                           x   x
                         AddV2(output)
        Args:
            input_graph_def: A GraphDef containing a model.

        Returns:
            Modified graph with individual ops that made up of layer normalization
            fused to LayerNorm.

        Raises:
            ValueError: If the graph is badly formed with duplicate node names.
        """
        input_node_map = {}
        for node in self.input_graph_def.node:
            if node.name not in input_node_map:
                input_node_map[node.name] = node
            else:
                raise ValueError("Duplicate node names detected for ", node.name)

        nodes_to_skip = {}
        new_ops = []
        for node in self.input_graph_def.node:
            if node.op != "AddV2":
                continue

            # AddV2 (Mul, beta) or AddV2 (beta, Mul)
            input0_op = node_from_map(input_node_map, node.input[0])
            input1_op = node_from_map(input_node_map, node.input[1])

            if input0_op.op == "Mul":
                data_scale_mul_op = input0_op
                beta_op = input1_op
            elif input1_op.op == "Mul":
                beta_op = input0_op
                data_scale_mul_op = input1_op
            else:
                continue

            # Mul (Reshape, *gamma)
            input0_op = node_from_map(input_node_map, data_scale_mul_op.input[0])
            input1_op = node_from_map(input_node_map, data_scale_mul_op.input[1])
            if input0_op.op == "Reshape":
                post_reshape_op = input0_op
                gamma_op = input1_op
            elif input1_op.op == "Reshape":
                post_reshape_op = input1_op
                gamma_op = input0_op
            else:
                continue

            # Reshape (FusedBatchNormV3, *post_shape)
            input0_op = node_from_map(input_node_map, post_reshape_op.input[0])
            input1_op = node_from_map(input_node_map, post_reshape_op.input[1])
            if input0_op.op == "FusedBatchNormV3":
                fused_batch_norm_op = input0_op
                post_shape_op = input1_op
            elif input1_op.op == "FusedBatchNormV3":
                fused_batch_norm_op = input1_op
                post_shape_op = input0_op
            else:
                continue

            # LayerNorm uses FusedBatchNorm in training mode.
            if fused_batch_norm_op.attr["is_training"] == attr_value_pb2.AttrValue(b=False):
                continue

            # FusedBatchNormV3(Reshape, Fill, Fill, Mean, Variance)
            pre_reshape_op = node_from_map(input_node_map, fused_batch_norm_op.input[0])
            if pre_reshape_op.op != "Reshape":
                continue
            fill_scale_op = node_from_map(input_node_map, fused_batch_norm_op.input[1])
            if fill_scale_op.op != "Fill":
                continue
            fill_offset_op = node_from_map(input_node_map, fused_batch_norm_op.input[2])
            if fill_offset_op.op != "Fill":
                continue

            # FusedBatchNorm node should have mean/variance as empty constant
            mean_op = node_from_map(input_node_map, fused_batch_norm_op.input[3])
            if mean_op.op != "Const":
                continue
            variance_op = node_from_map(input_node_map, fused_batch_norm_op.input[4])
            if variance_op.op != "Const":
                continue
            mean_value = values_from_const(mean_op)
            if mean_value.any():
                continue
            variance_value = values_from_const(variance_op)
            if variance_value.any():
                continue

            # Reshape (*input, *pre_shape)
            input_op = node_from_map(input_node_map, pre_reshape_op.input[0])
            pre_shape_op = node_from_map(input_node_map, pre_reshape_op.input[1])

            # Fill Scale(*dims_fill_scale, unit_gamma)
            dims_fill_scale_op = node_from_map(input_node_map, fill_scale_op.input[0])
            unit_gamma_op = node_from_map(input_node_map, fill_scale_op.input[1])
            if unit_gamma_op.op != "Const":
                continue

            # Fill Offset(*dims_fill_scale, unit_gamma)
            dims_fill_offset_op = node_from_map(input_node_map, fill_offset_op.input[0])
            zero_beta_op = node_from_map(input_node_map, fill_offset_op.input[1])
            if zero_beta_op.op != "Const":
                continue

            nodes_to_skip[node.name] = True
            nodes_to_skip[data_scale_mul_op.name] = True
            nodes_to_skip[post_reshape_op.name] = True
            nodes_to_skip[fused_batch_norm_op.name] = True
            nodes_to_skip[fill_scale_op.name] = True
            nodes_to_skip[fill_offset_op.name] = True

            new_fused_layernorm_op = node_def_pb2.NodeDef()
            new_fused_layernorm_op.op = "_MklLayerNorm"
            new_fused_layernorm_op.name = node.name
            new_fused_layernorm_op.attr["T"].CopyFrom(node.attr["T"])
            new_fused_layernorm_op.input.extend([input_op.name, gamma_op.name, beta_op.name])

            new_ops.append(new_fused_layernorm_op)

        result_graph_def = graph_pb2.GraphDef()
        for node in self.input_graph_def.node:
            if node.name in nodes_to_skip:
                continue
            new_node = node_def_pb2.NodeDef()
            new_node.CopyFrom(node)
            retained_input = []
            for input_node in new_node.input:
                if not input_node.startswith("^") or input_node[1:] not in nodes_to_skip:
                    retained_input.append(input_node)
            new_node.input[:] = retained_input
            result_graph_def.node.append(new_node)

        result_graph_def.node.extend(new_ops)
        result_graph_def.versions.CopyFrom(self.input_graph_def.versions)
        return result_graph_def


def node_name_from_input(node_name):  # pragma: no cover
    """Strips off ports and other decorations to get the underlying node name."""
    if node_name.startswith("^"):
        node_name = node_name[1:]
    m = re.search(r"(.*):\d+$", node_name)
    if m:
        node_name = m.group(1)
    return node_name


def node_from_map(node_map, name):  # pragma: no cover
    """Pulls a node def from a dictionary for a given name.

    Args:
        node_map: Dictionary containing an entry indexed by name for every node.
        name: Identifies the node we want to find.

    Returns:
        NodeDef of the node with the given name.

    Raises:
        ValueError: If the node isn't present in the dictionary.
    """
    stripped_name = node_name_from_input(name)
    if stripped_name not in node_map:
        raise ValueError("No node named '%s' found in map." % name)
    return node_map[stripped_name]


def values_from_const(node_def):  # pragma: no cover
    """Extracts the values from a const NodeDef as a numpy ndarray.

    Args:
        node_def: Const NodeDef that has the values we want to access.

    Returns:
        Numpy ndarray containing the values.

    Raises:
        ValueError: If the node isn't a Const.
    """
    if node_def.op != "Const":
        raise ValueError("Can not extract constant value from a node that is not Const. Got:\n" f"{node_def}")
    input_tensor = node_def.attr["value"].tensor
    tensor_value = tensor_util.MakeNdarray(input_tensor)
    return tensor_value



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Quantize ConcatV2 to int8 op."""

import re

from tensorflow.core.framework import node_def_pb2
from tensorflow.python.framework import dtypes

from neural_compressor.tensorflow.quantization.utils.quantize_graph_common import QuantizeGraphHelper as helper

from .quantize_graph_base import QuantizeNodeBase


class FuseNodeStartWithConcatV2(QuantizeNodeBase):
    """Quantize ConcatV2 to int8 op QuantizedConcatV2."""

    def _apply_concatv2_transform(self, original_node):
        """Quantize ConcatV2."""
        namespace_prefix = original_node.name + "_eightbit"
        quantized_concat_name = namespace_prefix + "_quantized_concatv2"
        reshape_dims_name, reduction_dims_name = self._add_common_quantization_nodes(
            namespace_prefix, helper.node_name_from_input(original_node.input[-1])
        )
        num_input = len(original_node.input)
        shape_input_name = original_node.input[num_input - 1]
        original_inputs = original_node.input[0 : num_input - 1]
        input_names = []
        min_names = []
        max_names = []
        for original_input_name in original_inputs:
            original_input_node = self.node_name_mapping[original_input_name].node
            input_data_type = dtypes.quint8 if self._find_relu_node(original_input_node) else dtypes.qint8
            quantize_input_name, min_input_name, max_input_name = self._eightbitize_input_to_node(
                namespace_prefix, original_input_name, reshape_dims_name, reduction_dims_name, dtype=input_data_type
            )
            input_names.append(quantize_input_name)
            min_names.append(min_input_name)
            max_names.append(max_input_name)
        all_input_names = input_names
        all_input_names.append(shape_input_name)
        all_input_names.extend(min_names)
        all_input_names.extend(max_names)
        quantized_concat_node = helper.create_node("QuantizedConcatV2", quantized_concat_name, all_input_names)
        helper.set_attr_int(quantized_concat_node, "N", len(original_inputs))
        helper.set_attr_dtype(quantized_concat_node, "T", input_data_type)
        self.add_output_graph_node(quantized_concat_node)
        self._intel_cpu_add_dequantize_result_node(quantized_concat_name, original_node.name, input_data_type)

    def _quantizable_concat(self, node):
        """Check if the ConcatV2 is quantizable."""
        deq_type = []
        for input_node_name in node.input[: node.attr["N"].i]:
            node_name = helper.node_name_from_input(input_node_name)
            if self.node_name_mapping[node_name].node.op != "Dequantize":
                return False

            deq_type.append(self.node_name_mapping[node_name].node.attr["T"].type)

        if len(set(deq_type)) != 1:
            return False

        return True

    def _apply_concatv2_quantization(self):
        """Quantize ConcatV2 if it's quantizable."""
        for _, v in self.node_name_mapping.items():
            if (
                v.node.op in ("ConcatV2",)
                and self._quantizable_concat(v.node)
                and dtypes.as_dtype(v.node.attr["T"].type) == dtypes.float32
                and not re.search(r"map(_\d+)?/while", v.node.name)
            ):
                self._apply_concatv2_transform(v.node)
                self.quantizable_node_names.append(v.node.name)
            else:
                new_node = node_def_pb2.NodeDef()
                new_node.CopyFrom(v.node)
                self.add_output_graph_node(new_node)

    def get_longest_fuse(self):
        """Only single ConcatV2, no fusion pattern."""
        return 1

    def apply_the_transform(self):
        """Apply the quantization of ConcatV2."""
        self.quantizable_node_names = []
        self._apply_concatv2_quantization()
        self._reset_output_node_maps()
        if self.remove_redundant_quant_flag:
            self.output_graph = self.remove_redundant_quantization(self.output_graph)
        return self.output_graph, self.quantizable_node_names, []



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""The core components of sq."""

from typing import Callable, Dict

import tensorflow as tf

from neural_compressor.common import logger
from neural_compressor.common.utils import DEFAULT_WORKSPACE
from neural_compressor.tensorflow.algorithms.smoother.calibration import (
    SmoothQuantCalibration,
    SmoothQuantCalibrationLLM,
)
from neural_compressor.tensorflow.algorithms.smoother.scaler import SmoothQuantScaler, SmoothQuantScalerLLM
from neural_compressor.tensorflow.quantization.config import SmoothQuantConfig
from neural_compressor.tensorflow.quantization.utils.graph_util import GraphAnalyzer
from neural_compressor.tensorflow.utils import SPR_BASE_VERSIONS, BaseModel, TensorflowLLMModel, TFConfig


class SmoothQuant:
    """The class that performs smooth quantization."""

    def __init__(
        self,
        config: SmoothQuantConfig,
        calib_dataloader: Callable = None,
        calib_iteration: int = 1,
        calib_func: Callable = None,
    ):
        """Convert the model by smooth quant.

        Args:
            config: the SmoothQuantConfig class used to set this class.
            calibdataloader: the calibration dataloader.
            calib_iteration: how many steps of iterations on the dataloader to move forward.
            calib_func: the function used for calibration, should be a substitution for calib_dataloader
            when the built-in calibration function of INC does not work for model inference.

        Returns:
            model: A smoothed Tensorflow model
        """
        assert calib_func is None, "calibration function is not supported for smooth quant."
        self.config = config
        self.calib_dataloader = calib_dataloader
        self.calib_iteration = calib_iteration

        self.new_api = tf.version.VERSION in SPR_BASE_VERSIONS
        self.device = TFConfig.global_config["device"]
        self.itex_mode = TFConfig.global_config["backend"] == "itex"

        for _, value in self.config.items():
            single_config = value
            break

        self.alpha = single_config.alpha
        self.folding = single_config.folding
        self.percentile = single_config.percentile
        self.op_types = single_config.op_types
        self.scales_per_op = single_config.scales_per_op
        self.record_max_info = single_config.record_max_info
        self.weight_clip = single_config.weight_clip
        self.auto_alpha_args = single_config.auto_alpha_args

    def get_weight_from_input_tensor(self, model, input_tensor_names):
        """Extracts weight tensors and their associated nodes from a smooth quant node's input tensor.

        Args:
            model: A TensorFlow model containing a `graph_def` attribute.
            input_tensor_names: A list of input tensor names to search for weight tensors.

        Returns:
            A tuple of two dictionaries:
            - sq_weight_tensors: A dictionary mapping each input tensor name
                to a dict of its associated weight tensors with weight name.
            - sq_weights_nodes: A dictionary mapping each input tensor name
                to a dict of its associated weight nodes with weight name.
        """
        g_analyzer = GraphAnalyzer()
        g_analyzer.graph = model.graph_def
        graph_info = g_analyzer.parse_graph()

        sq_weight_tensors = {}
        sq_weights_nodes = {}

        from tensorflow.python.framework import tensor_util

        for name in input_tensor_names:
            # Use dict rather than list to fix the QKV/VQK misorder issue
            curr_weight_tensors = {}
            curr_weights_nodes = {}
            next_node_names = graph_info[name].outputs
            for node_name in next_node_names:
                curr_node = graph_info[node_name].node
                if curr_node.op not in self.op_types:
                    continue
                if len(curr_node.input) >= 2:
                    weight_name = curr_node.input[1]
                    weight_node = graph_info[weight_name].node
                    weight_tensor = tensor_util.MakeNdarray(weight_node.attr["value"].tensor)
                    curr_weight_tensors[weight_name] = weight_tensor
                    curr_weights_nodes[weight_name] = weight_node
            # {input node -> {xxx_q_proj_matmul: value1, xxx_v_proj_matmul: value2, ...}, ...}
            sq_weight_tensors[name] = curr_weight_tensors
            sq_weights_nodes[name] = curr_weights_nodes
        return sq_weight_tensors, sq_weights_nodes

    def apply_smooth_quant(self, model: BaseModel):
        """Apply smooth quant to the model."""
        logger.info("Start Smoothing process for Smooth Quantization.")

        # Do a pre-optimization before smooth quant
        from neural_compressor.tensorflow.quantization.utils.graph_rewriter.generic.pre_optimize import PreOptimization

        pre_optimizer_handle = PreOptimization(model, self.new_api, self.device)
        pre_optimized_model = pre_optimizer_handle.get_optimized_model(self.itex_mode)
        model.graph_def = pre_optimized_model.graph_def

        # Run calibration to get max values per channel

        calibration = SmoothQuantCalibration(
            model, self.calib_dataloader, self.calib_iteration, self.op_types, self.percentile
        )
        max_vals_per_channel, sq_weight_node_names = calibration()

        # Get weight tensors and weight nodes based on the input tensor
        sq_weight_tensors, sq_weights_nodes = self.get_weight_from_input_tensor(model, max_vals_per_channel.keys())

        # Calculate the smooth quant scaler and insert Mul op into the graph
        scaler = SmoothQuantScaler(model, self.calib_dataloader, self.alpha, self.scales_per_op)
        model, mul_list = scaler.transform(
            max_vals_per_channel, sq_weight_tensors, sq_weights_nodes, sq_weight_node_names
        )

        return model

    def apply_smooth_quant_LLM(self, model: BaseModel):
        """Apply smooth quant to the LLM model."""
        # Do a pre-optimization before smooth quant
        from neural_compressor.tensorflow.quantization.utils.graph_rewriter.generic.pre_optimize import PreOptimization

        pre_optimizer_handle = PreOptimization(model, self.new_api, self.device)
        pre_optimized_model = pre_optimizer_handle.get_optimized_model(self.itex_mode)
        model.graph_def = pre_optimized_model.graph_def

        llm_temp_dir = DEFAULT_WORKSPACE + "/temp_saved_model"
        # Run calibration to get max values per channel
        calibration = SmoothQuantCalibrationLLM(
            model._model,
            self.calib_dataloader,
            self.calib_iteration,
            self.op_types,
            self.percentile,
            llm_temp_dir,
            model.weight_name_mapping,
        )
        max_vals_per_channel, sq_target_node_names, sq_weight_tensor_dict, sq_graph_def = calibration(
            model.input_node_names, model.output_node_names
        )

        # Calculate the smooth quant scaler and insert Mul op into the graph
        scaler = SmoothQuantScalerLLM(sq_graph_def, self.alpha, self.scales_per_op, self.op_types)
        sq_graph_def, sq_weight_scale_dict, mul_list = scaler.transform(
            max_vals_per_channel, sq_weight_tensor_dict, sq_target_node_names
        )
        model.graph_def = sq_graph_def
        model.model_path = llm_temp_dir
        model.sq_weight_scale_dict = sq_weight_scale_dict
        return model

    def __call__(self, model: BaseModel):
        """Convert the model by smooth quant.

        Args:
            model: original model

        Returns:
            model: A smoothed Tensorflow model
        """
        apply_func = self.apply_smooth_quant_LLM if isinstance(model, TensorflowLLMModel) else self.apply_smooth_quant

        return apply_func(model)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Initialize custom dense layers for Keras quantization."""

import json

import tensorflow as tf
from tensorflow.keras import activations, backend, constraints, initializers, regularizers
from tensorflow.keras.layers import Dense

from neural_compressor.tensorflow.utils import version1_gte_version2


class QDense(Dense):
    """The custom quantized Dense layer."""

    def __init__(
        self,
        name,
        units,
        activation=None,
        use_bias=True,
        kernel_initializer="glorot_uniform",
        bias_initializer="zeros",
        kernel_regularizer=None,
        bias_regularizer=None,
        activity_regularizer=None,
        kernel_constraint=None,
        bias_constraint=None,
        act_min_value=None,
        act_max_value=None,
        weight_min_value=None,
        weight_max_value=None,
        granularity="per_tensor",
        quant_status="calib",
        quant_mode="SCALED",
        quant_T="s8",
        quant_round_mode="HALF_AWAY_FROM_ZERO",
        quant_narrow_range=False,
        quant_axis=None,
        **kwargs
    ):
        """Initialize custom quantized Dense layer."""
        super(QDense, self).__init__(
            name=name,
            units=units,
            activation=activation,
            use_bias=use_bias,
            kernel_initializer=kernel_initializer,
            bias_initializer=bias_initializer,
            kernel_regularizer=kernel_regularizer,
            bias_regularizer=bias_regularizer,
            activity_regularizer=activity_regularizer,
            kernel_constraint=kernel_constraint,
            bias_constraint=bias_constraint,
            **kwargs
        )
        T_map = {"s8": tf.qint8, "u8": tf.quint8}
        self.weight_min_value = weight_min_value
        self.weight_max_value = weight_max_value
        self.act_min_value = act_min_value
        self.act_max_value = act_max_value
        self.granularity = granularity
        self.quant_status = quant_status
        self.quant_mode = quant_mode
        self.quant_T = T_map[quant_T]
        self.quant_round_mode = quant_round_mode
        self.quant_narrow_range = quant_narrow_range
        self.quant_axis = quant_axis

    def call(self, inputs):
        """The __call__ function of custom quantized Dense layer."""
        if self.quant_status == "calib" and not (
            version1_gte_version2(tf.__version__, "2.16.1") and isinstance(inputs, tf.keras.KerasTensor)
        ):
            if self.granularity == "per_tensor":
                self.act_min_value = tf.math.reduce_min(inputs)
                self.act_max_value = tf.math.reduce_max(inputs)
            else:
                self.act_min_value = tf.math.reduce_min(inputs, axis=1)
                self.act_max_value = tf.math.reduce_max(inputs, axis=1)
            kernel = self.kernel
        elif self.quant_status == "quantize":
            assert self.act_min_value is not None, "Invalid activation min-max values, please check calibration process"
            inputs, _, _ = tf.quantization.quantize(
                inputs,
                self.act_min_value,
                self.act_max_value,
                self.quant_T,
                mode=self.quant_mode,
                round_mode=self.quant_round_mode,
                narrow_range=self.quant_narrow_range,
                axis=self.quant_axis,
            )
            inputs = tf.quantization.dequantize(
                inputs,
                self.act_min_value,
                self.act_max_value,
                mode=self.quant_mode,
                narrow_range=self.quant_narrow_range,
                axis=self.quant_axis,
            )

            kernel_size = self.kernel.shape[-1]

            if not self.weight_min_value:
                self.weight_min_value = [-10000] * kernel_size
            if not self.weight_max_value:
                self.weight_max_value = [10000] * kernel_size

            # add the Q/DQ here
            kernel, _, _ = tf.quantization.quantize(
                self.kernel,
                self.weight_min_value,
                self.weight_max_value,
                tf.qint8,
                axis=1,
                mode="SCALED",
            )
            kernel = tf.quantization.dequantize(
                kernel,
                self.weight_min_value,
                self.weight_max_value,
                axis=1,
                mode="SCALED",
            )

        outputs = tf.keras.backend.dot(inputs, kernel)

        if self.use_bias:
            outputs = tf.keras.backend.bias_add(outputs, self.bias)
        if self.activation is not None:
            outputs = self.activation(outputs)
        return outputs

    @classmethod
    def from_config(cls, config):
        """Deserialize this class from a config dict."""
        return cls(**config)

    def get_config(self):
        """Serialize this class to a config dict."""
        config = super(QDense, self).get_config()
        config.update(
            {
                "act_min_value": self.act_min_value,
                "act_max_value": self.act_max_value,
                "weight_min_value": self.weight_min_value,
                "weight_max_value": self.weight_max_value,
                "granularity": self.granularity,
                "quant_status": self.quant_status,
                "quant_mode": self.quant_mode,
                "quant_T": "s8" if self.quant_T == tf.qint8 else "u8",
                "quant_round_mode": self.quant_round_mode,
                "quant_narrow_range": self.quant_narrow_range,
                "quant_axis": self.quant_axis,
            }
        )

        return config


def initialize_int8_dense(fp32_layer, q_config):
    """Initialize int8 dense."""
    kwargs = fp32_layer.get_config()

    param_list = [
        "name",
        "units",
        "activation",
        "use_bias",
        "kernel_initializer",
        "bias_initializer",
        "kernel_regularizer",
        "activity_regularizer",
        "bias_regularizer",
        "kernel_constraint",
        "bias_constraint",
    ]
    for p in param_list:  # pragma: no cover
        if p in kwargs:
            del kwargs[p]

    q_layer = QDense(
        name=fp32_layer.name,
        units=fp32_layer.units,
        activation=fp32_layer.activation,
        use_bias=fp32_layer.use_bias,
        kernel_initializer=fp32_layer.kernel_initializer,
        bias_initializer=fp32_layer.bias_initializer,
        kernel_regularizer=fp32_layer.kernel_regularizer,
        bias_regularizer=fp32_layer.bias_regularizer,
        activity_regularizer=fp32_layer.activity_regularizer,
        kernel_constraint=fp32_layer.kernel_constraint,
        bias_constraint=fp32_layer.bias_constraint,
        quant_T=q_config["T"],
        granularity=q_config["granularity"],
        **kwargs
    )

    return q_layer



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Freeze FakeQuant op Graph Rewriter."""

from tensorflow.python.framework import dtypes

from neural_compressor.tensorflow.quantization.utils.graph_util import GraphAnalyzer, GraphRewriterHelper
from neural_compressor.tensorflow.utils import dump_elapsed_time

from ..graph_base import GraphRewriterBase


class FreezeFakeQuantOpOptimizer(GraphRewriterBase):  # pragma: no cover
    """Freeze fake_quant op to the following Quantize op and prioring Dequantize op."""

    def __init__(self, model):
        """Initialization."""
        super().__init__(model)

        self.graph_analyzer = GraphAnalyzer()
        self.graph_analyzer.graph = self.model

        self.graph_info = self.graph_analyzer.parse_graph()

        self.freeze_patterns = {
            str(
                [["Requantize", "RequantizePerChannel"], ["Dequantize"], ["FakeQuantWithMinMaxVars"]]
            ): self._freeze_requant_dequant_fakequant,
            str([["FakeQuantWithMinMaxVars"], ["QuantizeV2"]]): self._freeze_fakequant_quant,
            str(
                [["FakeQuantWithMinMaxVars"], ["Shape"], ["StridedSlice"], ["Pack"], ["Reshape"], ["QuantizeV2"]]
            ): self._freeze_fakequant_metaop_quant,
        }

    def _freeze_requant_dequant_fakequant(self, pattern_nodes):
        """Freeze Requantize Dequantize FakeQuant fusion."""
        requant_node_name = pattern_nodes[0]
        requant_node = self.graph_info[requant_node_name].node
        fake_quant_node_name = pattern_nodes[2]
        fake_quant_node = self.graph_info[fake_quant_node_name].node

        # set the third input "requested_output_min" of RequantizePerChannel or Requantize op.
        requested_output_min_node = self.graph_info[requant_node.input[3]].node
        min_node = self.graph_info[fake_quant_node.input[1]].node
        GraphRewriterHelper.set_attr_tensor(
            requested_output_min_node, "value", min_node.attr["value"].tensor, dtypes.float32
        )
        # set the fourth input "requested_output_max" of RequantizePerChannel or Requantize op.
        requested_output_max_node = self.graph_info[requant_node.input[4]].node
        max_node = self.graph_info[fake_quant_node.input[2]].node
        GraphRewriterHelper.set_attr_tensor(
            requested_output_max_node, "value", max_node.attr["value"].tensor, dtypes.float32
        )

    def _freeze_fakequant_quant(self, pattern_nodes):
        """Freeze FakeQuant QuantizeV2 fusion."""
        fake_quant_node_name = pattern_nodes[0]
        fake_quant_node = self.graph_info[fake_quant_node_name].node
        quant_node_name = pattern_nodes[1]
        quant_node = self.graph_info[quant_node_name].node

        # set the second input "min_range" of QuantizeV2 op.
        min_node = self.graph_info[fake_quant_node.input[1]].node
        min_range_node = self.graph_info[quant_node.input[1]].node
        GraphRewriterHelper.set_attr_tensor(min_range_node, "value", min_node.attr["value"].tensor, dtypes.float32)
        # set the third input "max_range" of QuantizeV2 op.
        max_node = self.graph_info[fake_quant_node.input[2]].node
        max_range_node = self.graph_info[quant_node.input[2]].node
        GraphRewriterHelper.set_attr_tensor(max_range_node, "value", max_node.attr["value"].tensor, dtypes.float32)

    def _freeze_fakequant_metaop_quant(self, pattern_nodes):
        """Freeze FakeQuant Meta ops QuantizeV2 fusion."""
        fake_quant_node_name = pattern_nodes[0]
        fake_quant_node = self.graph_info[fake_quant_node_name].node
        quant_node_name = pattern_nodes[5]
        quant_node = self.graph_info[quant_node_name].node

        # set the second input "min_range" of QuantizeV2 op.
        min_node = self.graph_info[fake_quant_node.input[1]].node
        min_range_node = self.graph_info[quant_node.input[1]].node
        GraphRewriterHelper.set_attr_tensor(min_range_node, "value", min_node.attr["value"].tensor, dtypes.float32)
        # set the third input "max_range" of QuantizeV2 op.
        max_node = self.graph_info[fake_quant_node.input[2]].node
        max_range_node = self.graph_info[quant_node.input[2]].node
        GraphRewriterHelper.set_attr_tensor(max_range_node, "value", max_node.attr["value"].tensor, dtypes.float32)

    def _remove_all_fake_quants(self):
        """Remove all the fake quants."""
        _const_node = []

        for node_name in list(self.graph_info.keys()):
            node = self.graph_info[node_name].node
            if node.op == "FakeQuantWithMinMaxVars":
                origin_outputs = list(self.graph_info[node_name].outputs)
                min_node_name = self.graph_info[node.input[1]].node.name
                max_node_name = self.graph_info[node.input[2]].node.name
                _const_node.append(min_node_name)
                _const_node.append(max_node_name)

                self.graph_analyzer.remove_node_with_single_input_output(node_name)

                for j in origin_outputs[1:]:
                    output_node = self.graph_info[j].node
                    if (
                        len(output_node.input) == 1
                        and output_node.op == "Const"
                        and output_node.input[0] == "^" + node.name
                    ):
                        self.graph_info[j].node.ClearField("input")
                    elif output_node.op == "NoOp":
                        new_noop_input = [
                            noop_input for noop_input in output_node.input if noop_input != "^" + node.name
                        ]
                        output_node.ClearField("input")
                        output_node.input.extend(new_noop_input)

        # remove those left const nodes used by FakeQuantWithMinMaxVars
        for node_name in list(self.graph_info.keys()):
            if node_name in _const_node:
                self.graph_analyzer.remove_node(node_name)

    @dump_elapsed_time("Pass FreezeFakeQuantOpOptimizer")
    def do_transformation(self):
        """Execute freeze FakeQuant optimization."""
        for _pattern, _handler in self.freeze_patterns.items():
            _match_pattern_nodes = self.graph_analyzer.query_fusion_pattern_nodes(eval(_pattern))
            if len(_match_pattern_nodes) == 0:
                continue

            for _pattern_nodes in _match_pattern_nodes:
                _handler(_pattern_nodes)

        self._remove_all_fake_quants()

        return GraphAnalyzer().dump_graph()
